#!/usr/bin/env python
"""
_GetTask_

Oracle implementation of Jobs.Task
"""

__all__ = []



import logging

from WMCore.WMBS.MySQL.Jobs.GetTask import GetTask as MySQLGetTask

class GetTask(MySQLGetTask):
    """
    Identical to MySQL version
    """



#!/usr/bin/env python
"""
_FailOrphanFiles_

MySQL implementation of Subscription.FailOrphanFiles
"""

from WMCore.Database.DBFormatter import DBFormatter

class FailOrphanFiles(DBFormatter):
    sql = """SELECT DISTINCT avail_files.fileid FROM
               (SELECT wmbs_sub_files_available.fileid AS fileid, :fileset AS fileset
                       FROM wmbs_sub_files_available
                WHERE wmbs_sub_files_available.subscription = :subscription) avail_files
               INNER JOIN wmbs_file_parent ON
                 avail_files.fileid = wmbs_file_parent.child
               INNER JOIN wmbs_job_assoc ON
                 wmbs_file_parent.parent = wmbs_job_assoc.fileid
               INNER JOIN wmbs_workflow_output ON
                 avail_files.fileset = wmbs_workflow_output.output_fileset
               INNER JOIN wmbs_subscription ON
                 wmbs_workflow_output.workflow_id = wmbs_subscription.workflow
               INNER JOIN wmbs_jobgroup ON
                 wmbs_subscription.id = wmbs_jobgroup.subscription
               INNER JOIN wmbs_job ON
                 wmbs_job.jobgroup = wmbs_jobgroup.id AND
                 wmbs_job_assoc.job = wmbs_job.id AND
                 wmbs_job.outcome = 0"""

    sqlFail = """INSERT INTO wmbs_sub_files_failed (subscription, fileid)
                   VALUES (:subscription, :fileid)"""

    sqlAvail = """DELETE FROM wmbs_sub_files_available
                  WHERE subscription = :subscription AND
                        fileid = :fileid"""

    def execute(self, subscriptionID, filesetID, conn = None, transaction = False):
        binds = {"subscription": subscriptionID, "fileset": filesetID}
        results = self.dbi.processData(self.sql, binds, conn = conn,
                                       transaction = transaction)

        binds = []
        for result in self.formatDict(results):
            binds.append({"subscription": subscriptionID,
                          "fileid": result["fileid"]})

        if len(binds) > 0:
            self.dbi.processData(self.sqlAvail, binds, conn = conn,
                                 transaction = transaction)
            self.dbi.processData(self.sqlFail, binds, conn = conn,
                                 transaction = transaction)
        return



#!/usr/bin/env python
"""
_LoadFromID_

Oracle implementation of Subscription.LoadFromID
"""




from WMCore.WMBS.MySQL.Subscriptions.LoadFromID import LoadFromID as LoadFromIDMySQL

class LoadFromID(LoadFromIDMySQL):
    pass



#!/usr/bin/env python
"""
_Exists_

Oracle implementation of WorkflowExists

"""
__all__ = []



from WMCore.WMBS.MySQL.Workflow.Exists import Exists as WorkflowExistsMySQL

class Exists(WorkflowExistsMySQL):
    sql = WorkflowExistsMySQL.sql



#!/usr/bin/env python
"""
_couch_archiver_

Replicate an agent's jobdump to another couch machine.
"""
from __future__ import print_function

import os
import sys

from WMCore.Configuration import loadConfigurationFile
from WMCore.Database.CMSCouch import CouchServer
from WMCore.Lexicon import sanitizeURL

if len(sys.argv) != 3:
    print("Usage: ")
    print("  archiver.py [Destination Couch URL] [Destination DB Name]")
    print("")
    print("Example:")
    print("  archiver.py http://username:password@cms-xen41.fnal.gov:5984 cmssrv98_070_jobdump")
    sys.exit(0)

wmagentConfig = loadConfigurationFile(os.environ["WMAGENT_CONFIG"])
srcCouchHost = wmagentConfig.JobStateMachine.couchurl
srcDbBase = wmagentConfig.JobStateMachine.couchDBName
destCouchHost = sys.argv[1]
destDbBase = sys.argv[2]

destCouchServer = CouchServer(dburl = destCouchHost)
srcCouchServer = CouchServer(dburl = srcCouchHost)

srcJobsDb = srcCouchHost + "/" + srcDbBase + "%2Fjobs"
destJobsDb = destCouchHost + "/" + destDbBase + "%2Fjobs"
srcFwjrsDb = srcCouchHost + "/" + srcDbBase + "%2Ffwjrs"
destFwjrsDb = destCouchHost + "/" + destDbBase + "%2Ffwjrs"

print("Archiving %s/%s to %s/%s..." % (srcCouchHost, srcDbBase, destCouchHost, destDbBase))

# Replicate the FWJR and Jobs databases...
print("  Replicating jobs database...")
destCouchServer.replicate(srcJobsDb, destJobsDb, create_target = True)
print("  Replication fwjrs database...")
destCouchServer.replicate(srcFwjrsDb, destFwjrsDb, create_target = True)

# Generate views for the various databases
destJobsDb = destCouchServer.connectDatabase(destDbBase + "/jobs")
destFwjrsDb = destCouchServer.connectDatabase(destDbBase + "/fwjrs")
print("  Triggering view generation for jobs database...")
destJobsDb.loadView("JobDump", "statusByWorkflowName", options = {"limit": 1})
print("  Triggering view generation for fwjrs database...")
destFwjrsDb.loadView("FWJRDump", "outputByWorkflowName", options = {"limit": 1})

print("")
# Query destination DB for list of workflows
summaryBase = "%s/%s%%2Ffwjrs/_design/FWJRDump/_show/workflowSummary/%s" # dest host, dest db base, workflow name
successBase = "%s/%s%%2Fjobs/_design/JobDump/_list/successJobs/statusByWorkflowName?startkey=%%5B%%22%s%%22%%5D&endkey=%%5B%%22%s%%22%%2C%%7B%%7D%%5D&reduce=false" # dest host, dest db base, workflow, workflow
failedBase = "%s/%s%%2Fjobs/_design/JobDump/_list/failedJobs/statusByWorkflowName?startkey=%%5B%%22%s%%22%%5D&endkey=%%5B%%22%s%%22%%2C%%7B%%7D%%5D&reduce=false" # dest host, dest db base, workflow, workflow

srcJobsDb = srcCouchServer.connectDatabase(srcDbBase + "/jobs")
statusResult = srcJobsDb.loadView("JobDump", "statusByWorkflowName", options = {"group_level": 1})

fileHandle = open("archived.html", "w")
fileHandle.write("<html><head><title>Archived Workflows</title></head>\n")
fileHandle.write("<body>\n")

workflowNames = []
for statusRow in statusResult["rows"]:
    wfName = statusRow["key"][0]
    summaryUrl = summaryBase % (destCouchHost, destDbBase, wfName)
    successUrl = successBase % (destCouchHost, destDbBase, wfName, wfName)
    failedUrl = successBase % (destCouchHost, destDbBase, wfName, wfName)
    fileHandle.write("%s " % wfName)
    fileHandle.write("<a href=%s>(summary)</a>" % sanitizeURL(summaryUrl)["url"])
    fileHandle.write(" <a href=%s>(success)</a>" % sanitizeURL(successUrl)["url"])
    fileHandle.write(" <a href=%s>(failure)</a><br>\n" % sanitizeURL(failedUrl)["url"])

fileHandle.write("</body></html>\n")
fileHandle.close()



"""
_WorkUnit.ExistsByTaskFileLumi_

Oracle implementation of WorkUnit.ExistsByTaskFileLumi
"""

from __future__ import absolute_import, division, print_function

from WMCore.WMBS.MySQL.WorkUnit.ExistsByTaskFileLumi import ExistsByTaskFileLumi as ExistsByTaskFileLumiMySQL


class ExistsByTaskFileLumi(ExistsByTaskFileLumiMySQL):
    """
    _WorkUnit.ExistsByTaskFileLumi_

    Oracle implementation of WorkUnit.ExistsByTaskFileLumi
    """

    pass



from __future__ import division, print_function

DEFAULT_AGENT_CONFIG = {
    # user drain mode which is decided by user
    "UserDrainMode": False,
    # agent drain mode automatically calculated by agent
    "AgentDrainMode": False,
    # speed drain flag used for informational purposes only (reported back to wmstats)
    "SpeedDrainMode": False,
    # disk usage percentage for warning
    "DiskUseThreshold": 85,
     # list of disks which shouldn't be included in monitoring for the Threshold
    "IgnoreDisks": ["/mnt/ramdisk"],
    # fraction of condor schedd limit, used for job submission
    "CondorJobsFraction": 0.75,
    # In JobSumitter, submit jobs over the threshold if the priority is higher than current pending/running jobs
    "CondorOverflowFraction": 0.2,  # default 20% over threshold
    # ExitCodes returned by jobs which doesn't need to be retried
    "NoRetryExitCodes": [70, 73, 8001, 8006, 8009, 8023, 8026, 8501, 50660, 50661, 50664, 71102, 71104, 71105],
    # Number of times a job will be retried
    "MaxRetries" : {'default': 3, 'Merge': 4, 'Cleanup': 2, 'LogCollect': 1, 'Harvesting': 4},
    # Speed drain thresholds - a draining agent's config is updated when # of remaining condor jobs is below a threshold
    # CondorPriority: increase condor priority for all queued Production/Processing jobs to 999999
    # NoJobRetries: change MaxRetries to 0, all job errors are terminal
    # EnableAllSites: submit jobs to any site
    "SpeedDrainConfig": {'CondorPriority': {'Threshold': 500, 'Enabled': False},
                         'NoJobRetries':   {'Threshold': 200, 'Enabled': False},
                         'EnableAllSites': {'Threshold': 200, 'Enabled': False}
                        }
}



#!/usr/bin/env python
"""
_PSetTweaks_

Record a set of tweakable parameters from a CMSSW Configuration in a CMSSW
independent python structure

"""

from future import standard_library
standard_library.install_aliases()

from builtins import object, map, range
from past.builtins import basestring
from future.utils import viewitems, viewvalues

import imp
import inspect
import json
import pickle
import sys
from functools import reduce

class PSetHolder(object):
    """
    _PSetHolder_

    Dummy PSet object used to construct the Tweak object to mimic
    the config structure

    """
    def __init__(self, psetName):
        self.psetName_ = psetName
        self.parameters_ = []

#  //
# // Assistant lambda functions
#//
childPSets = lambda x: [ value for value in viewvalues(x.__dict__)
                         if value.__class__.__name__ == "PSetHolder" ]
childParameters = lambda p, x: [ "%s.%s" % (p,i) for i in  x.parameters_ ]

recursiveGetattr = lambda obj, attr: reduce(getattr, attr.split("."), obj)


def parameterIterator(obj):
    """
    _parameterIterator_

    Util to iterate through the parameters in a PSetHolder

    """
    x = None
    for x in childParameters(obj):
        yield getattr(obj, x)

def psetIterator(obj):
    """
    _psetIterator_

    Util to iterate through the child psets in a PSetHolder
    """
    for x in childPSets(obj):
        yield x




class PSetLister(object):
    """
    _PSetLister_

    Operator to decompose the PSet structure into a
    more sequence like get up

    """
    def __init__(self):
        self.psets = []
        self.parameters = {}
        self.queue = []


    def __call__(self, pset):
        """
        _operator(PSetHolder)_

        recursively traverse all parameters in this and all child
        PSets

        """
        self.queue.append(pset.psetName_)
        psetPath = ".".join(self.queue)
        self.psets.append(psetPath)
        params = childParameters(psetPath, pset)
        self.parameters[psetPath] = params
        list(map(self, childPSets(pset)))
        self.queue.pop(-1)


class JSONiser(object):
    """
    _JSONiser_

    Util class to build a json dictionary structure from the PSet tree
    and also recover the pset tree from a json structure

    """
    def __init__(self):
        self.json = {}
        self.queue = []
        self.parameters = {}


    def __call__(self, pset, parent = None):
        """
        _operator(pset)_

        operate on pset and substructure to build a json dictionary

        """
        if parent == None: parent = self.json

        thisPSet = parent.get(pset.psetName_, None)
        if thisPSet == None:
            parent[pset.psetName_] = {}
            thisPSet = parent[pset.psetName_]

        for param in pset.parameters_:
            thisPSet[param] = getattr(pset, param)
        thisPSet['parameters_'] = pset.parameters_

        for child in childPSets(pset):
            self(child, thisPSet)


    def dejson(self, dictionary):
        """
        _dejson_

        Convert the json structure back to PSetHolders

        """
        params = dictionary.get('parameters_', [])
        queue = ".".join(self.queue)
        for param in params:
            self.parameters["%s.%s" % (queue, param)]  = dictionary[param]
        for key, value in viewitems(dictionary):
            if isinstance(value, dict):
                self.queue.append(key)
                self.dejson(dictionary[key])
                self.queue.pop(-1)








class PSetTweak(object):
    """
    _PSetTweak_

    Template object listing the parameters to be edited.
    Also provides serialisation functionality and defines the
    process + tweak operator to apply the tweaks to a process.

    """

    def __init__(self):
        self.process = PSetHolder("process")

    def addParameter(self, attrName, value):
        """
        _addAttribute_

        Add an attribute as process.pset1.pset2.param = value
        Value should be the appropriate python type

        """
        currentPSet = None
        paramList = attrName.split(".")
        for _ in range(0, len(paramList)):
            param = paramList.pop(0)
            if param == "process":
                currentPSet = self.process
            elif len(paramList) > 0:
                if not hasattr(currentPSet, param):
                    setattr(currentPSet, param, PSetHolder(param))
                currentPSet = getattr(currentPSet, param)
            else:
                setattr(currentPSet, param, value)
                currentPSet.parameters_.append(param)


    def getParameter(self, paramName):
        """
        _getParameter_

        Get value of the parameter with the name given of the
        form process.module...

        """
        if not paramName.startswith("process"):
            msg = "Invalid Parameter Name: %s\n" % paramName
            msg += "Parameter must start with process"
            raise RuntimeError(msg)
        return recursiveGetattr(self, paramName)


    def __iter__(self):
        """
        _iterate_

        Loop over all parameters in the tweak, returning the
        parameter name as a . delimited path and the value

        """
        lister = PSetLister()
        lister(self.process)

        for pset in lister.psets:
            for param in lister.parameters[pset]:
                yield param , self.getParameter(param)

    def psets(self):
        """
        _psets_

        Generator function to yield the PSets in the tweak

        """
        lister = PSetLister()
        lister(self.process)

        for pset in lister.psets:
            yield pset






    def __str__(self):
        """string repr for debugging etc"""
        result = ""
        for x,y in self:
            result += "%s = %s\n" % (x, y)
        return result


    def setattrCalls(self, psetPath):
        """
        _setattrCalls_

        Generate setattr call for each parameter in the pset structure
        Used for generating python format

        """
        result = {}
        current = None
        last = None
        psets = psetPath.split(".")
        for _ in range(0, len(psets)):
            pset = psets.pop(0)
            last = current
            if current == None:
                current = pset
            else:
                current += ".%s" % pset
            if last != None:
                result[current] = "setattr(%s, \"%s\", PSetHolder(\"%s\"))" % (
                    last, pset, pset)
        return result



    def pythonise(self):
        """
        _pythonise_

        return this object as python format

        """
        src = inspect.getsourcelines(PSetHolder)
        result = ""
        for line in src[0]:
            result += line

        result += "\n\n"
        result += "# define PSet Structure\n"
        result += "process = PSetHolder(\"process\")\n"
        setattrCalls = {}
        for pset in self.psets():
            setattrCalls.update(self.setattrCalls(pset))
        order = sorted(setattrCalls.keys())
        for call in order:
            if call == "process": continue
            result += "%s\n" % setattrCalls[call]

        result += "# set parameters\n"
        for param, value in self:
            psetName = param.rsplit(".", 1)[0]
            paramName = param.rsplit(".", 1)[1]
            if isinstance(value, basestring):
                value = "\"%s\"" % value
            result += "setattr(%s, \"%s\", %s)\n" % (
                psetName, paramName, value)
            result += "%s.parameters_.append(\"%s\")\n" % (psetName, paramName)

        return result


    def jsonise(self):
        """
        _jsonise_

        return json format of this tweak

        """
        jsoniser = JSONiser()
        jsoniser(self.process)
        result = json.dumps(jsoniser.json)
        return result

    def jsondictionary(self):
        """
        _jsondictionary_

        return the json layout dictionary, rather than stringing it

        """
        jsoniser = JSONiser()
        jsoniser(self.process)
        return jsoniser.json

    def simplejsonise(self):
        """
        _simplejsonise_

        return simple json format of this tweak
        E.g.:
        {"process.maxEvents.input": 1200, "process.source.firstRun": 1, "process.source.firstLuminosityBlock": 59965}

        """
        jsoniser = JSONiser()
        jsoniser.dejson(self.jsondictionary())
        result = json.dumps(jsoniser.parameters)
        return result


    def persist(self, filename, formatting="python"):
        """
        _persist_

        Save this object as either python, json or pickle

        """
        if formatting not in ("python", "json", "pickle", "simplejson"):
            msg = "Unsupported Format: %s" % formatting
            raise RuntimeError(msg)

        if formatting == "python":
            with open(filename, 'w') as handle:
                handle.write(self.pythonise())
        if formatting == "json":
            with open(filename, "w") as handle:
                handle.write(self.jsonise())
        if formatting == "pickle":
            with open(filename, "wb") as handle:
                pickle.dump(self, handle)
        if formatting == "simplejson":
            with open(filename, "w") as handle:
                handle.write(self.simplejsonise())
        return

    def unpersist(self, filename, formatting=None):
        """
        _unpersist_

        Load data from file provided, if format is not specified, guess
        it based on file extension

        """
        if formatting == None:
            fileSuffix = filename.rsplit(".", 1)[1]
            if fileSuffix == "py":
                formatting = "python"
            if fileSuffix == "pkl":
                formatting = "pickle"
            if fileSuffix == "json":
                formatting = "json"

        if formatting not in ("python", "json", "pickle"):
            msg = "Unsupported Format: %s" % formatting
            raise RuntimeError(msg)

        if formatting == "pickle":
            with open(filename, 'rb') as handle:
                unpickle = pickle.load(handle)
            self.process.__dict__.update(unpickle.__dict__)

        if formatting == "python":
            modRef = imp.load_source('tempTweak', filename)
            lister = PSetLister()
            lister(modRef.process)
            for pset in lister.psets:
                for param in lister.parameters[pset]:
                    self.addParameter(param , recursiveGetattr(modRef, param))
            del modRef, sys.modules['tempTweak']


        if formatting == "json":
            with open(filename, 'r') as handle:
                jsonContent = handle.read()


            jsoniser = JSONiser()
            jsoniser.dejson(json.loads(jsonContent))

            for param, value in viewitems(jsoniser.parameters):
                self.addParameter(param , value)

    def reset(self):
        """
        _reset_

        Reset pset holder process

        """
        del self.process
        self.process = PSetHolder("process")


def makeTweakFromJSON(jsonDictionary):
    """
    _makeTweakFromJSON_

    Make a tweak instance and populate it from a dictionary JSON
    structure

    """
    jsoniser = JSONiser()
    jsoniser.dejson(jsonDictionary)
    tweak = PSetTweak()
    for param, value in viewitems(jsoniser.parameters):
        tweak.addParameter(param , value)
    return tweak



#!/usr/bin/env python
"""
_WMBSBase_

Generic methods used by all of the WMBS classes.
"""




from WMCore.WMConnectionBase import WMConnectionBase

class WMBSBase(WMConnectionBase):
    """
    Generic methods used by all of the WMBS classes.
    """
    def __init__(self):
        """
        ___init___

        Initialize all the database connection attributes and the logging
        attritbutes.  Create a DAO factory for WMCore.WMBS as well. Finally,
        check to see if a transaction object has been created.  If none exists,
        create one but leave the transaction closed.
        """
        WMConnectionBase.__init__(self, daoPackage = "WMCore.WMBS")



"""
_WorkUnit.GetByTaskFileLumi_

Oracle implementation of WorkUnit.GetByTaskFileLumi
"""

from __future__ import absolute_import, division, print_function

from WMCore.WMBS.MySQL.WorkUnit.GetByTaskFileLumi import GetByTaskFileLumi as GetByTaskFileLumiMySQL


class GetByTaskFileLumi(GetByTaskFileLumiMySQL):
    """
    _WorkUnit.GetByTaskFileLumi_

    Oracle implementation of WorkUnit.GetByTaskFileLumi
    """

    pass



#!/usr/bin/env python
"""
_HeritageLFNChild_

Oracle implementation of DBSBufferFiles.HeritageLFNChild
"""




from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.HeritageLFNChild import \
     HeritageLFNChild as MySQLHeritageLFNChild

class HeritageLFNChild(MySQLHeritageLFNChild):
    pass



from __future__ import print_function, division
from WMCore.Database.DBFormatter import DBFormatter

class CheckStatusForCompletedWorkflows(DBFormatter):
    """
    _CheckStatusForCompletedWorkflows_

    Retrieve information about the workflow which checks the status of PhEDEx and DBS injection.
    There is case task is completed but no file is associated to it.(EmptyBlock flag)
    Not sure how that happens - It could be all failed case, or task is not related to producing files
    """
    sql = """SELECT dw.name, df.in_phedex, df.status, count(*) as count
                FROM dbsbuffer_workflow dw  
                INNER JOIN wmbs_workflow ww on ww.name = dw.name
                LEFT JOIN dbsbuffer_file df on dw.id = df.workflow
             WHERE dw.completed = 1 group by dw.name, df.in_phedex, df.status """

    def format(self, results):
        results = self.formatDict(results)
        resultByWorkflow = {}
        for info in results:
            workflow = info['name']
            resultByWorkflow.setdefault(workflow, {})
            resultByWorkflow[workflow].setdefault("InDBS", 0)
            resultByWorkflow[workflow].setdefault("NotInDBS", 0)
            resultByWorkflow[workflow].setdefault("InPhEDEx", 0)
            resultByWorkflow[workflow].setdefault("NotInPhEDEx", 0)
            resultByWorkflow[workflow].setdefault("NoNeedToUpload", 0)

            count = info["count"]

            if info['status'] == "InDBS":
                resultByWorkflow[workflow]["InDBS"] += count
            elif info['status'] == "NOTUPLOADED":
                resultByWorkflow[workflow]["NotInDBS"] += count

            if info['status'] == "InDBS" or info['status'] == "NOTUPLOADED":
                if info['in_phedex'] == 1:
                    resultByWorkflow[workflow]["InPhEDEx"] += count
                elif info['in_phedex'] == 0:
                    resultByWorkflow[workflow]["NotInPhEDEx"] += count
            else:
                # This contains the case
                # info['status'] == GLOBAL - which means all the files are input files which already in DBS
                # info['status'] is None -- which means there is no associate file for the task. Maybe all are failed
                resultByWorkflow[workflow]["NoNeedToUpload"] += count

        return resultByWorkflow
            
        
    def execute(self, conn=None, transaction=False):
        """
        _execute_

        Retrieve information about a workflow status association in DBSBuffer.
        """
        results = self.dbi.processData(self.sql, conn = conn,
                                          transaction = transaction)

        return self.format(results)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Simple in-memory and non-thread safe cache.
Note that this module does not support home-made object types, since there is
an explicit data type check when adding a new item to the cache.

It raises a TypeError exception if the cache data type chagens;
or if the user tries to extend the cache with an incompatible
data type.
"""

from __future__ import (print_function, division)

from copy import copy

from builtins import object
from time import time


class MemoryCacheException(Exception):
    def __init__(self, message):
        super(MemoryCacheException, self).__init__(message)


class MemoryCache(object):

    __slots__ = ["lastUpdate", "expiration", "_cache"]

    def __init__(self, expiration, initialData=None):
        """
        Initializes cache object

        :param expiration: expiration time in seconds
        :param initialData: initial value for the cache
        """
        self.lastUpdate = int(time())
        self.expiration = expiration
        self._cache = initialData

    def __contains__(self, item):
        """
        Check whether item is in the current cache
        :param item: a simple object (string, integer, etc)
        :return: True if the object can be found in the cache, False otherwise
        """
        return item in self._cache

    def __getitem__(self, keyName):
        """
        If the cache is a dictionary, return that item from the cache. Else, raise an exception.
        :param keyName: the key name from the dictionary
        """
        if isinstance(self._cache, dict):
            return copy(self._cache.get(keyName))
        else:
            raise MemoryCacheException("Cannot retrieve an item from a non-dict MemoryCache object: {}".format(self._cache))

    def reset(self):
        """
        Resets the cache to its current data type
        """
        if isinstance(self._cache, (dict, set)):
            self._cache.clear()
        elif isinstance(self._cache, list):
            del self._cache[:]
        else:
            raise MemoryCacheException("The cache needs to be reset manually, data type unknown")

    def isCacheExpired(self):
        """
        Evaluate whether the cache has already expired, returning
        True if it did, otherwise it returns False
        """
        return self.lastUpdate + self.expiration < int(time())

    def getCache(self):
        """
        Raises an exception if the cache has expired, otherwise returns
        its data
        """
        if self.isCacheExpired():
            expiredSince = int(time()) - (self.lastUpdate + self.expiration)
            raise MemoryCacheException("Memory cache expired for %d seconds" % expiredSince)
        return self._cache

    def setCache(self, inputData):
        """
        Refresh the cache with the content provided (refresh its expiration as well)
        This method enforces the user to not change the cache data type
        :param inputData: data to store in the cache
        """
        if not isinstance(self._cache, type(inputData)):
            raise TypeError("Current cache data type: %s, while new value is: %s" %
                            (type(self._cache), type(inputData)))
        self.reset()
        self.lastUpdate = int(time())
        self._cache = inputData

    def addItemToCache(self, inputItem):
        """
        Adds new item(s) to the cache, without resetting its expiration.
        It, of course, only works for data caches of type: list, set or dict.
        :param inputItem: additional item to be added to the current cached data
        """
        if isinstance(self._cache, set) and isinstance(inputItem, (list, set)):
            # extend another list or set into a set
            self._cache.update(inputItem)
        elif isinstance(self._cache, set) and isinstance(inputItem, (int, float, str)):
            # add a simple object (integer, string, etc) to a set
            self._cache.add(inputItem)
        elif isinstance(self._cache, list) and isinstance(inputItem, (list, set)):
            # extend another list or set into a list
            self._cache.extend(inputItem)
        elif isinstance(self._cache, list) and isinstance(inputItem, (int, float, str)):
            # add a simple object (integer, string, etc) to a list
            self._cache.append(inputItem)
        elif isinstance(self._cache, dict) and isinstance(inputItem, dict):
            self._cache.update(inputItem)
        else:
            msg = "Input item type: %s cannot be added to a cache type: %s" % (type(self._cache), type(inputItem))
            raise TypeError("Cache and input item data type mismatch. %s" % msg)



"""
Oracle implementation of AddRunLumiFile
"""
from WMCore.WMBS.MySQL.Files.AddRunLumi import AddRunLumi as AddRunLumiMySQL

class AddRunLumi(AddRunLumiMySQL):
    """
    _AddRunLumi_

    overwirtes MySQL Files.AddRunLumi.sql to use in oracle.

    To Check: might not need to overwrite if modifiy MySQL.Files.AddRunLumi.sql
    a bit to work with both
    """
    sql = """INSERT INTO wmbs_file_runlumi_map (fileid, run, lumi, num_events)
                SELECT wfd.id, :run, :lumi, :num_events FROM wmbs_file_details wfd WHERE lfn = :lfn
                  AND NOT EXISTS (SELECT fileid FROM wmbs_file_runlumi_map wfrm2
                                   WHERE wfrm2.fileid = wfd.id
                                   AND wfrm2.run = :run
                                   AND wfrm2.lumi = :lumi)"""



#!/usr/bin/env python
"""
_GetAvailableFiles_

Oracle implementation of Subscription.GetAvailableFiles
"""

from WMCore.WMBS.MySQL.Subscriptions.GetAvailableFiles import \
     GetAvailableFiles as GetAvailableFilesMySQL

class GetAvailableFiles(GetAvailableFilesMySQL):
    pass



#!/usr/bin/env python
"""
_GetLocation_

MySQL implementation of File.GetLocation
"""

from WMCore.Database.DBFormatter import DBFormatter


class GetLocation(DBFormatter):
    sql = """SELECT DISTINCT wpnn.pnn AS pnn FROM wmbs_file_location wfl
               INNER JOIN wmbs_file_details wfd ON wfd.id = wfl.fileid
               INNER JOIN wmbs_pnns wpnn ON wfl.pnn = wpnn.id
               WHERE wfd.lfn = :lfn"""

    def getBinds(self, file=None):
        binds = []
        file = self.dbi.makelist(file)
        for f in file:
            binds.append({'lfn': f})
        return binds

    def format(self, result):
        "Return a list of SE FQDN's"
        out = set()
        for r in result:
            for i in r.fetchall():
                out.add(i[0])
            r.close()
        return out

    def execute(self, file=None, conn=None, transaction=False):
        binds = self.getBinds(file)

        result = self.dbi.processData(self.sql, binds,
                                      conn=conn, transaction=transaction)
        return self.format(result)



#!/usr/bin/env python
"""
_SetLocation_

MySQL implementation of Jobs.SetLocation
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class SetLocation(DBFormatter):
    """
    _GetLocation_

    Retrieve all files that are associated with the given job from the
    database.
    """
    sql = """UPDATE wmbs_job SET location = (SELECT id FROM wmbs_location WHERE site_name = :location)
              WHERE id = :jobid"""


    def execute(self, jobid = None, location = None, bulkList = None, conn = None, transaction = False):
        """
        _execute_

        Execute the SQL for the given job ID and location
        """


        if bulkList:
            binds = bulkList
        else:
            binds = {'jobid': jobid, 'location': location}

        self.dbi.processData(self.sql, binds, conn = conn, transaction = transaction)
        return



#!/usr/bin/env python
"""
_GetNumberOfJobsPerSite_

MySQL implementation of Jobs.GetNumberOfJobsPerSite
"""

__all__ = []



import logging

from WMCore.Database.DBFormatter import DBFormatter

class GetNumberOfJobsPerSite(DBFormatter):
    """
    _GetLocation_

    Retrieve all files that are associated with the given job from the
    database.
    """

    def buildSQL(self, states, type):
        """
        _buildSQL_

        builds the sql statements; necessary for lists
        """


        baseSQL = """SELECT count(*) FROM wmbs_job
              WHERE location = (SELECT ID FROM wmbs_location WHERE site_name = :location)
              AND state IN (SELECT ID FROM wmbs_job_state js WHERE js.name IN (
        """

        typeSQL = """SELECT count(*) FROM wmbs_job
                       INNER JOIN wmbs_jobgroup ON wmbs_job.jobgroup = wmbs_jobgroup.id
                       INNER JOIN wmbs_subscription ON wmbs_jobgroup.subscription = wmbs_subscription.id
                       INNER JOIN wmbs_job_state ON wmbs_job.state = wmbs_job_state.id
                       INNER JOIN wmbs_location ON wmbs_job.location = wmbs_location.id
                       INNER JOIN wmbs_sub_types ON wmbs_subscription.subtype = wmbs_sub_types.id
                       WHERE wmbs_location.site_name = :location
                       AND wmbs_sub_types.name = :type
                       AND wmbs_job_state.name IN ("""

        if type:
            sql = typeSQL
        else:
            sql = baseSQL

        states = list(states)

        count = 0
        for state in states:
            if not count == 0:
                sql += ", "
            sql += ":state%i" %(count)
            count += 1
        sql += ")"

        return sql


    def format(self, results):
        """
        _format_

        """

        if len(results) == 0:
            return False
        else:
            return results[0].fetchall()[0]


    def buildBinds(self, location, states, type):
        """

        _buildBinds_

        Build a list of binds

        """

        binds = {}

        binds['location']     = location
        count = 0
        for state in states:
            binds["state%i" %(count)] = state
            count += 1

        if type:
            binds['type'] = type


        return binds


    def execute(self, location, states, type = None, conn = None, transaction = False):
        """
        _execute_

        Execute the SQL for the given job ID and then format and return
        the result.
        """

        sql = self.buildSQL(states, type)

        binds = self.buildBinds(location, states, type)

        #print "In Jobs.GetNumberOfJobsPerSite"
        #print sql
        #print binds

        result = self.dbi.processData(sql, binds, conn = conn, transaction = transaction)


        return self.format(result)



"""
Easily get the version of the python interpreter at runtime
"""

from __future__ import division # Jenkins CI

import sys
import pickle

PY3 = sys.version_info[0] == 3
PY2 = sys.version_info[0] == 2

# We need to keep compatibility between multiple python versions
# Further details at: https://github.com/dmwm/WMCore/pull/10726
# For PY2: set highest protocol to 2
# For PY3: set highest protocol to 4 (compatible with python3.6 and python3.8)
HIGHEST_PICKLE_PROTOCOL = pickle.HIGHEST_PROTOCOL if PY2 else 4


#!/usr/bin/env python
"""
_Lexicon_

A set of regular expressions  and other tests that we can use to validate input
to other classes. If a test fails an AssertionError should be raised, and
handled appropriately by the client methods, on success returns True.
"""
from __future__ import print_function, division

from future import standard_library
standard_library.install_aliases()
from builtins import str, bytes
from future.utils import viewvalues, viewkeys

import io
import re
import json

from urllib.parse import urlparse, urlunparse

from WMCore.WMException import WMException, WMEXCEPTION_START_STR, WMEXCEPTION_END_STR

# restriction enforced by DBS. for different types blocks.
# It could have a strict restriction
# i.e production should end with v[number]
PRIMARY_DS = {'re': '^[a-zA-Z][a-zA-Z0-9\-_]*$', 'maxLength': 99}
PROCESSED_DS = {'re': '[a-zA-Z0-9\.\-_]+', 'maxLength': 199}
TIER = {'re': '[A-Z\-_]+', 'maxLength': 99}
BLOCK_STR = {'re': '#[a-zA-Z0-9\.\-_]+', 'maxLength': 100}

lfnParts = {
    'era': '([a-zA-Z0-9\-_]+)',
    'primDS': '([a-zA-Z][a-zA-Z0-9\-_]*)',
    'tier': '(%(re)s)' % TIER,
    'version': '([a-zA-Z0-9\-_]+)',
    'procDS': '([a-zA-Z0-9\-_]+)',  # Processed dataset = Processing string + Processing version
    'counter': '([0-9]+)',
    'root': '([a-zA-Z0-9\-_]+).root',
    'hnName': '([a-zA-Z0-9\.]+)',
    'subdir': '([a-zA-Z0-9\-_]+)',
    'file': '([a-zA-Z0-9\-\._]+)',
    'workflow': '([a-zA-Z0-9\-_]+)',
    'physics_group': '([a-zA-Z0-9\-_]+)'
}

userProcDSParts = {
    'groupuser': '([a-zA-Z0-9\.\-_])+',
    'publishdataname': '([a-zA-Z0-9\-_])+',
    'psethash': '([a-f0-9]){32}'
}

STORE_RESULTS_LFN = '/store/results/%(physics_group)s/%(era)s/%(primDS)s/%(tier)s/%(procDS)s' % lfnParts

# condor log filtering lexicons
WMEXCEPTION_FILTER = "(?P<WMException>\%s(?!<@).*?\%s)" % (WMEXCEPTION_START_STR, WMEXCEPTION_END_STR)
WMEXCEPTION_FILTER += "|(?P<ERROR>(ERROR:root:.*?This is a CRITICAL error))"
WMEXCEPTION_REGEXP = re.compile(r"%s" % WMEXCEPTION_FILTER, re.DOTALL)

CONDOR_LOG_REASON_FILTER = '<a n="Reason"><s>(?P<Reason>(?!</s></a>).*?)</s></a>'
CONDOR_LOG_SITE_FILTER = '<a n="MachineAttrGLIDEIN_CMSSite0"><s>(?P<Site>(?!</s></a>).*?)</s></a>'

CONDOR_LOG_FILTER_REGEXP = re.compile(r"%s|%s" % (CONDOR_LOG_REASON_FILTER, CONDOR_LOG_SITE_FILTER),
                                      re.DOTALL)


def DBSUser(candidate):
    """
    create_by and last_modified_by in DBS are in several formats. The major ones are:
    1. DN that was mostly used in DBS2: example /DC=org/DC=doegrids/OU=People/CN=Lothar A.T. Bauerdick 301799;
    2. CERN HN account name that used in DBS3/CMSWEB if the HN is assocated with DN: example giffels ;
    3. username with host name: example cmsprod@vocms39.cern.ch;
    """
    if candidate == '' or not candidate:
        return candidate
    r1 = r'^/[a-zA-Z][a-zA-Z0-9/\=\s()\']*\=[a-zA-Z0-9/\=\.\-_/#:\s\']*$'
    r2 = r'^[a-zA-Z0-9/][a-zA-Z0-9/\.\-_\']*$'
    r3 = r'^[a-zA-Z0-9/][a-zA-Z0-9/\.\-_]*@[a-zA-Z0-9/][a-zA-Z0-9/\.\-_]*$'

    errorMsg = "DBSUser candidate: %s doesn't match any of the following regular expressions:\n" % candidate
    try:
        return check(r1, candidate)
    except AssertionError:
        errorMsg += "  %s\n" % r1

    try:
        return check(r2, candidate)
    except AssertionError:
        errorMsg += "  %s\n" % r2

    try:
        return check(r3, candidate)
    except AssertionError:
        errorMsg += "  %s\n" % r3
        raise AssertionError(errorMsg)


def searchblock(candidate):
    """
    A block name with a * wildcard one or more times in it.
    """
    regexp = r"^/(\*|[a-zA-Z\*][a-zA-Z0-9_\*]{0,100})(/(\*|[a-zA-Z0-9_\.\-\*]{1,199})){0,1}(/(\*|[A-Z\-\*]{1,99})(#(\*|[a-zA-Z0-9\.\-_\*]){0,100}){0,1}){0,1}$"
    return check(regexp, candidate)


SEARCHDATASET_RE = r'^/(\*|[a-zA-Z\*][a-zA-Z0-9_\*\-]{0,100})(/(\*|[a-zA-Z0-9_\.\-\*]{1,199})){0,1}(/(\*|[A-Z\-\*]{1,50})){0,1}$'


def searchdataset(candidate):
    """
    A dataset name with a * wildcard one or more times in it. Only the first '/' is mandatory to use.
    """
    return check(SEARCHDATASET_RE, candidate)


def searchstr(candidate):
    """
    Used to check a DBS input that searches for names in dbs. Note block name, dataset name, file name have their own
    searching string.
    No white space found in names in DBS production and allowed to elimate input like "Drop table table1".
    letters, numbers, periods, dashes, underscores

    """
    if candidate == '':
        return candidate
    return check(r'^[a-zA-Z0-9/%*][a-zA-Z0-9/\.\-_%*/#]*$', candidate)


def namestr(candidate):
    """
    Any input used in DBS and not defined here should pass namestr check.
    No white space found in names in DBS production and allowed to elimate input like "Drop table table1".
    letters, numbers, periods, dashes, underscores,#,/

    """
    if candidate == '' or not candidate:
        return candidate
    return check(r'^[a-zA-Z0-9/][a-zA-Z0-9/\.\-_/#]*$', candidate)


def sitetier(candidate):
    return check("^T[0-3]", candidate)


def jobrange(candidate):
    """ Specifies a numbers/range of jobs separated by a comma.
        A range is composed by two numbers separated my minus
        For example valid candidates are either 1 or 1,2 or 3-6,5,7-8
        It is like when you specifies which pages to print in Word
    """
    return check("^\d+(-\d+)?(,\d+(-\d+)?)*$", candidate)


def cmsname(candidate):
    """
    Candidate must pass the CMS name pattern. Thus:
     * good candidates: T2_UK_SGrid or T2_UK_SGrid_Bristol
     * bad candidates: T2, T2_UK
    """
    candidate = candidate.rstrip('_')
    return check("^T[0-3]_[A-Z]{2}((_[A-Za-z0-9]+)+)$", candidate)


def countrycode(candidate):
    # TODO: do properly with a look up table
    return check("^[A-Z]{2}$", candidate)


def _blockStructCheck(candidate):
    """
    Basic block structure check
    /primary/process/tier#uuid
    """
    assert candidate.count('/') == 3, "need to have / between the 3 parts which construct block name"
    parts = candidate.split('/')
    assert parts[3].count('#') == 1, "need to have # in the last parts of block"
    # should be empty string for the first part
    check(r"", parts[0])
    return parts


def block(candidate):
    """assert if not a valid block name"""

    parts = _blockStructCheck(candidate)

    primDSCheck = check(r"%s" % PRIMARY_DS['re'], parts[1], PRIMARY_DS['maxLength'])
    procDSCheck = check(r"%s" % PROCESSED_DS['re'], parts[2], PROCESSED_DS['maxLength'])
    lastParts = parts[3].split("#")
    tierCheck = check(r"%s" % TIER['re'], lastParts[0], TIER['maxLength'])
    blockCheck = check(r"%s" % BLOCK_STR['re'], "#%s" % lastParts[1], BLOCK_STR['maxLength'])
    return (primDSCheck and procDSCheck and tierCheck and blockCheck)


def identifier(candidate):
    """ letters, numbers, whitespace, periods, dashes, underscores """
    return check(r'[a-zA-Z0-9\s\.\-_]{1,100}$', candidate)


def globalTag(candidate):
    """ Identifier plus colons """
    return check(r'[a-zA-Z0-9\s\.\-_:]{1,100}$', candidate)


DATASET_RE = r'^/[a-zA-Z0-9\-_]{1,99}/[a-zA-Z0-9\.\-_]{1,199}/[A-Z\-]{1,50}$'


def dataset(candidate):
    """ A slash followed by an identifier,x3 """
    return check(DATASET_RE, candidate)


PROCDATASET_RE = r'[a-zA-Z][a-zA-Z0-9_]*(\-[a-zA-Z0-9_]+){0,2}-v[0-9]*$'


def procdataset(candidate):
    """
    Check for processed dataset name.
    letters, numbers, dashes, underscores.
    """
    if not candidate or candidate.startswith('None'):
        raise AssertionError("ProcDataset cannot be empty or start with None.")

    commonCheck = check(r"%s" % PROCESSED_DS['re'], candidate, PROCESSED_DS['maxLength'])
    prodCheck = check(PROCDATASET_RE, candidate)
    return (commonCheck and prodCheck)


def publishdatasetname(candidate):
    if candidate == '' or not candidate:
        return candidate
    return check(r'%(publishdataname)s$' % userProcDSParts, candidate, 100)


USERPROCDATASET_RE = r'%(groupuser)s-%(publishdataname)s-%(psethash)s$' % userProcDSParts


def userprocdataset(candidate):
    """
    Check for processed dataset name of users.
    letters, numbers, dashes, underscores.
    """
    if candidate == '' or not candidate:
        return candidate

    commonCheck = check(r"%s" % PROCESSED_DS['re'], candidate, PROCESSED_DS['maxLength'])
    anlaysisCheck = check(USERPROCDATASET_RE, candidate)
    return (commonCheck and anlaysisCheck)


def physicsgroup(candidate):
    """
    Check for Physics Group string which is added to StoreResults
    merged LFN base. Up to 30 letters, numbers, dashes, underscores.
    """
    return check(r'%(physics_group)s$' % lfnParts, candidate, 30)


def procversion(candidate):
    """ Integers """
    if isinstance(candidate, dict):
        for candi in viewvalues(candidate):
            check(r'^[0-9]+$', str(candi))
        return True
    else:
        return check(r'^[0-9]+$', str(candidate))


def procstring(candidate):
    """ Identifier """
    if not candidate:
        raise AssertionError("ProcStr cannot be empty or None.")
    if isinstance(candidate, dict):
        for candi in viewvalues(candidate):
            check(r'[a-zA-Z0-9_]{1,100}$', candi)
        return True
    else:
        return check(r'[a-zA-Z0-9_]{1,100}$', candidate)


def procstringT0(candidate):
    """
    ProcessingString validation function for T0 specs
    """
    if isinstance(candidate, dict):
        for candi in viewvalues(candidate):
            check(r'^$|[a-zA-Z0-9_]{1,100}$', candi)
        return True
    else:
        return check(r'^$|[a-zA-Z0-9_]{1,100}$', candidate)


def acqname(candidate):
    """
    Check for acquisition name.
    letters, numbers, underscores.
    """
    if not candidate:
        raise AssertionError("AcqEra cannot be empty or None.")
    if isinstance(candidate, dict):
        for candi in viewvalues(candidate):
            check(r'[a-zA-Z][a-zA-Z0-9_]*$', candi)
        return True
    else:
        return check(r'[a-zA-Z][a-zA-Z0-9_]*$', candidate)


def campaign(candidate):
    """
    Check for Campaign name.
    letters, numbers, underscores and dashes are allowed, up to 60 chars.
    """
    if not candidate:
        return True
    return check(r'^[a-zA-Z0-9-_]{1,80}$', candidate)


def primdataset(candidate):
    """
    Check for primary dataset name.
    letters, numbers, dashes, underscores.
    """
    if candidate == '' or not candidate:
        return candidate
    return check(r"%s" % PRIMARY_DS['re'], candidate, PRIMARY_DS['maxLength'])


TASK_STEP_NAME = {'re': '^[a-zA-Z][a-zA-Z0-9\-_]*$', 'maxLength': 50}
def taskStepName(candidate):
    """
    Validate the TaskName and/or StepName field.
    Letters, numbers, dashes and underscores are allowed.
    """
    return check(r"%s" % TASK_STEP_NAME['re'], candidate, TASK_STEP_NAME['maxLength'])

def hnName(candidate):
    """
    Use lfn parts definitions to validate a simple HN name
    """

    validName = '^%(hnName)s$' % lfnParts
    return check(validName, candidate)


def lfn(candidate):
    """
    Should be of the following form:
    /store/data/acquisition_era/primary-dataset/data_tier/processing_version/lfn_counter/filename.root
    See https://twiki.cern.ch/twiki/bin/viewauth/CMS/DMWMPG_Namespace for details

    NOTE:Because of the way we do lustre, we have to have two separate checks for this:
    /store/data
    /store/data/lustre

    Add for LHE files: /data/lhe/...
    """
    regexp1 = '/([a-z]+)/([a-z0-9]+)/(%(era)s)/([a-zA-Z0-9\-_]+)/([A-Z\-_]+)/([a-zA-Z0-9\-_]+)((/[0-9]+){3}){0,1}/([0-9]+)/([a-zA-Z0-9\-_]+).root' % lfnParts
    regexp2 = '/([a-z]+)/([a-z0-9]+)/([a-z0-9]+)/([a-zA-Z0-9\-_]+)/([a-zA-Z0-9\-_]+)/([A-Z\-_]+)/([a-zA-Z0-9\-_]+)((/[0-9]+){3}){0,1}/([0-9]+)/([a-zA-Z0-9\-_]+).root'
    regexp3 = '/store/(temp/)*(user|group)/(%(hnName)s|%(physics_group)s)/%(primDS)s/%(procDS)s/%(version)s/%(counter)s/%(root)s' % lfnParts
    regexp4 = '/store/(temp/)*(user|group)/(%(hnName)s|%(physics_group)s)/%(primDS)s/(%(subdir)s/)+%(root)s' % lfnParts

    oldStyleTier0LFN = '/store/data/%(era)s/%(primDS)s/%(tier)s/%(version)s/%(counter)s/%(counter)s/%(counter)s/%(root)s' % lfnParts
    tier0LFN = '/store/(backfill/[0-9]/){0,1}(t0temp/|unmerged/){0,1}(data|express|hidata)/%(era)s/%(primDS)s/%(tier)s/%(version)s/%(counter)s/%(counter)s/%(counter)s(/%(counter)s)?/%(root)s' % lfnParts

    storeMcLFN = '/store/mc/(%(era)s)/([a-zA-Z0-9\-_]+)/([a-zA-Z0-9\-_]+)/([a-zA-Z0-9\-_]+)(/([a-zA-Z0-9\-_]+))*/([a-zA-Z0-9\-_]+).root' % lfnParts

    storeResults2LFN = '/store/results/%(physics_group)s/%(primDS)s/%(procDS)s/%(primDS)s/%(tier)s/%(procDS)s/%(counter)s/%(root)s' % lfnParts

    storeResultRootPart = '%(counter)s/%(root)s' % lfnParts
    storeResultsLFN = "%s/%s" % (STORE_RESULTS_LFN, storeResultRootPart)

    lheLFN1 = '/store/lhe/([0-9]+)/([a-zA-Z0-9\-_]+).lhe(.xz){0,1}'
    # This is for future lhe LFN structure. Need to be tested.
    lheLFN2 = '/store/lhe/%(era)s/%(primDS)s/([0-9]+)/([a-zA-Z0-9\-_]+).lhe(.xz){0,1}' % lfnParts

    errorMsg = "LFN candidate: %s doesn't match any of the following regular expressions:\n" % candidate

    try:
        return check(regexp1, candidate)
    except AssertionError:
        errorMsg += "  %s\n" % regexp1

    try:
        return check(regexp2, candidate)
    except AssertionError:
        errorMsg += "  %s\n" % regexp2

    try:
        return check(regexp3, candidate)
    except AssertionError:
        errorMsg += "  %s\n" % regexp3

    try:
        return check(regexp4, candidate)
    except AssertionError:
        errorMsg += "  %s\n" % regexp4

    try:
        return check(tier0LFN, candidate)
    except AssertionError:
        errorMsg += "  %s\n" % tier0LFN

    try:
        return check(oldStyleTier0LFN, candidate)
    except AssertionError:
        errorMsg += "  %s\n" % oldStyleTier0LFN

    try:
        return check(storeMcLFN, candidate)
    except AssertionError:
        errorMsg += "  %s\n" % storeMcLFN

    try:
        return check(lheLFN1, candidate)
    except AssertionError:
        errorMsg += "  %s\n" % lheLFN1

    try:
        return check(lheLFN2, candidate)
    except AssertionError:
        errorMsg += "  %s\n" % lheLFN2

    try:
        return check(storeResults2LFN, candidate)
    except AssertionError:
        errorMsg += "  %s\n" % storeResults2LFN

    try:
        return check(storeResultsLFN, candidate)
    except AssertionError:
        errorMsg += "  %s\n" % storeResultsLFN
        raise AssertionError(errorMsg)


def lfnBase(candidate):
    """
    As lfn above, but for doing the lfnBase
    i.e., for use in spec generation and parsing
    """
    regexp1 = '/([a-z]+)/([a-z0-9]+)/([a-zA-Z0-9\-_]+)/([a-zA-Z0-9\-_]+)/([A-Z\-_]+)/([a-zA-Z0-9\-_]+)'
    regexp2 = '/([a-z]+)/([a-z0-9]+)/([a-z0-9]+)/([a-zA-Z0-9\-_]+)/([a-zA-Z0-9\-_]+)/([A-Z\-_]+)/([a-zA-Z0-9\-_]+)((/[0-9]+){3}){0,1}'
    regexp3 = '/(store)/(temp/)*(user|group)/(%(hnName)s|%(physics_group)s)/%(primDS)s/%(procDS)s/%(version)s' % lfnParts

    tier0LFN = '/store/(backfill/[0-9]/){0,1}(t0temp/|unmerged/){0,1}(data|express|hidata)/%(era)s/%(primDS)s/%(tier)s/%(version)s/%(counter)s/%(counter)s/%(counter)s' % lfnParts

    errorMsg = "LFN candidate: %s doesn't match any of the following regular expressions:\n" % candidate

    try:
        return check(regexp1, candidate)
    except AssertionError:
        errorMsg += "  %s\n" % regexp1

    try:
        return check(regexp2, candidate)
    except AssertionError:
        errorMsg += "  %s\n" % regexp2

    try:
        return check(regexp3, candidate)
    except AssertionError:
        errorMsg += "  %s\n" % regexp3

    try:
        return check(tier0LFN, candidate)
    except AssertionError:
        errorMsg += "  %s\n" % tier0LFN

    try:
        return check(STORE_RESULTS_LFN, candidate)
    except AssertionError:
        errorMsg += "  %s\n" % STORE_RESULTS_LFN
        raise AssertionError(errorMsg)


def userLfn(candidate):
    """
    Check LFNs in /store/{temp}/user that are not EDM data
    """
    regexp = '/store/(temp/)*(user|group)/(%(hnName)s|%(physics_group)s)/%(subdir)s/%(workflow)s/%(subdir)s/%(file)s' % lfnParts
    return check(regexp, candidate)


def userLfnBase(candidate):
    """
    As above but for the base part of the file
    """
    regexp = '/store/(temp/)*(user|group)/(%(hnName)s|%(physics_group)s)/%(subdir)s/%(workflow)s/%(subdir)s' % lfnParts
    return check(regexp, candidate)


def cmsswversion(candidate):
    return check('CMSSW(_\d+){3}(_[a-zA-Z0-9_]+)?$', candidate)


def couchurl(candidate):
    return check('https?://(([a-zA-Z0-9:@\.\-_]){0,100})([a-z0-9\.]+)(:\d+|/couchdb)', candidate)


def requestName(candidate):
    return check(r'[a-zA-Z0-9\.\-_]{1,150}$', candidate)


def validateUrl(candidate):
    """
    Basic input validation for http(s) urls
    """
    # regex taken from django https://github.com/django/django/blob/master/django/core/validators.py#L47
    # Copyright (c) Django Software Foundation and individual contributors
    protocol = r"^https?://"  # http:// or https://
    domain = r'?:(?:[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?\.)+[a-zA-Z]{2,6}\.?'
    localhost = r'localhost'
    ipv4 = r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}'
    ipv6 = r'\[?[a-fA-F0-9]*:[a-fA-F0-9:]+\]?'
    port = r'(?::\d+)?'  # optional port
    path = r'(?:/?|[/?]\S+)$'
    regex_url = r'%s(%s|%s|%s|%s)%s%s' % (protocol, domain, localhost, ipv4, ipv6, port, path)
    return check(regex_url, candidate)


def check(regexp, candidate, maxLength=None):
    if maxLength is not None:
        assert len(candidate) <= maxLength, \
            "%s is longer than max length (%s) allowed" % (candidate, maxLength)
    assert re.compile(regexp).match(candidate) is not None, \
        "'%s' does not match regular expression %s" % (candidate, regexp)
    return True


def parseLFN(candidate):
    """
    _parseLFN_

    Take an LFN, return the component parts
    """
    separator = "/"

    # First, make sure what we've gotten is a real LFN
    lfn(candidate)

    parts = candidate.split('/')
    final = {}

    if parts[0] == '':
        parts.remove('')
    if 'user' in parts[1:3] or 'group' in parts[1:3]:
        if parts[1] in ['user', 'group']:
            final['baseLocation'] = '/%s' % separator.join(parts[:2])
            parts = parts[2:]
        else:
            final['baseLocation'] = '/%s' % separator.join(parts[:3])
            parts = parts[3:]

        final['hnName'] = parts[0]
        final['primaryDataset'] = parts[1]
        final['secondaryDataset'] = parts[2]
        final['processingVersion'] = parts[3]
        final['lfnCounter'] = parts[4]
        final['filename'] = parts[5]

        return final

    if len(parts) == 8:
        # Then we have only two locations
        final['baseLocation'] = '/%s' % separator.join(parts[:2])
        parts = parts[2:]
    elif len(parts) == 9:
        final['baseLocation'] = '/%s' % separator.join(parts[:3])
        parts = parts[3:]
    else:
        # How did we end up here?
        # Something just went wrong
        msg = """CRITICAL!  This machine has experienced a complete logic failure while parsing LFNs.\n
        If you are a developer this indicates that you have changed the Lexicon LFN regexp functions without changing the parsing.\n
        If you are an operator, this indicates that this machine is likely unstable.\n
        All data should be backed up and the machine removed from production for examination.\n"""
        msg += "Candidate: %s" % candidate
        raise WMException(msg)

    final['acquisitionEra'] = parts[0]
    final['primaryDataset'] = parts[1]
    final['dataTier'] = parts[2]
    final['processingVersion'] = parts[3]
    final['lfnCounter'] = parts[4]
    final['filename'] = parts[5]

    return final


def parseLFNBase(candidate):
    """
    _parseLFNBase_

    Return a meaningful dictionary with info from an LFNBase
    """
    separator = "/"

    # First, make sure what we've gotten is a real LFNBase
    lfnBase(candidate)

    parts = candidate.split('/')
    final = {}

    if parts[0] == '':
        parts.remove('')

    if 'user' in parts[1:3] or 'group' in parts[1:3]:
        if parts[1] in ['user', 'group']:
            final['baseLocation'] = '/%s' % separator.join(parts[:2])
            parts = parts[2:]
        else:
            final['baseLocation'] = '/%s' % separator.join(parts[:3])
            parts = parts[3:]

        final['hnName'] = parts[0]
        final['primaryDataset'] = parts[1]
        final['secondaryDataset'] = parts[2]
        final['processingVersion'] = parts[3]

        return final

    if len(parts) == 6:
        # Then we have only two locations
        final['baseLocation'] = '/%s' % separator.join(parts[:2])
        parts = parts[2:]
    elif len(parts) == 7:
        final['baseLocation'] = '/%s' % separator.join(parts[:3])
        parts = parts[3:]
    else:
        # How did we end up here?
        # Something just went wrong
        msg = """CRITICAL!  This machine has experienced a complete logic failure while parsing LFNs.\n
        If you are a developer this indicates that you have changed the Lexicon LFN regexp functions without changing the parsing.\n
        If you are an operator, this indicates that this machine is likely unstable.\n
        All data should be backed up and the machine removed from production for examination.\n"""
        msg += "Candidate: %s" % candidate
        raise WMException(msg)

    final['acquisitionEra'] = parts[0]
    final['primaryDataset'] = parts[1]
    final['dataTier'] = parts[2]
    final['processingVersion'] = parts[3]

    return final


def sanitizeURL(url):
    """Take the url with/without username and password and return sanitized url,
       username and password in dict format
       WANNING: This doesn't check the correctness of url format.
       Don't use ':' in username or password.
    """
    endpoint_components = urlparse(url)
    # Cleanly pull out the user/password from the url
    if endpoint_components.port:
        netloc = '%s:%s' % (endpoint_components.hostname,
                            endpoint_components.port)
    else:
        netloc = endpoint_components.hostname

    # Build a URL without the username/password information
    url = urlunparse(
        [endpoint_components.scheme,
         netloc,
         endpoint_components.path,
         endpoint_components.params,
         endpoint_components.query,
         endpoint_components.fragment])

    return {'url': url, 'username': endpoint_components.username,
            'password': endpoint_components.password}


def replaceToSantizeURL(url_str):
    """
    Take arbitrary string and search for urls with user and password and
    replace it with sanitized url.
    """

    def _repUrl(matchObj):
        return matchObj.group(1) + matchObj.group(4)

    # TODO: won't catch every case (But is it good enough (trade off to performance)?)
    urlRegExpr = r'\b(((?i)http|https|ftp|mysql|oracle|sqlite)+://)([^:]+:[^@]+@)(\S+)\b'
    return re.sub(urlRegExpr, _repUrl, url_str)


def splitCouchServiceURL(serviceURL):
    """
    split service URL to couchURL and couchdb name
    serviceURL should be couchURL/dbname format.
    """

    splitedURL = serviceURL.rstrip('/').rsplit('/', 1)
    return splitedURL[0], splitedURL[1]


def primaryDatasetType(candidate):
    pDatasetTypes = ["mc", "data", "cosmic", "test"]
    if candidate in pDatasetTypes:
        return True
    # to sync with the check() exception when it doesn't match
    raise AssertionError("Invalid primary dataset type : %s should be 'mc' or 'data' or 'test'" % candidate)


def subRequestType(candidate):
    subTypes = ["MC", "ReDigi", "Pilot", "RelVal", "HIRelVal", "ReReco", ""]
    if candidate in subTypes:
        return True
    # to sync with the check() exception when it doesn't match
    msg = "Invalid SubRequestType value: '{}'. Allowed values are: {}".format(candidate, subTypes)
    raise AssertionError(msg)


def activity(candidate):
    dashboardActivities = ['reprocessing', 'production', 'relval', 'tier0', 't0',
                           'harvesting', 'storeresults', 'integration',
                           'test', 'analysis']
    if candidate in dashboardActivities:
        return True
    raise AssertionError("Invalid dashboard activity: %s should 'test'" % candidate)


def gpuParameters(candidate):
    """
    Validate the spec "GPUParams" argument, which is a JSON encoded object, thus:
    * an encoded None object (like 'null')
    * an encoded dictionary with the following parameters:
      * mandatory: GPUMemoryMB (int), CUDARuntime (str), CUDACapabilities (list of str)
      * optional: GPUName (str), CUDADriverVersion (str), CUDARuntimeVersion (str)
    :param candidate: a JSON encoded data to be validated
    :return: True if validation succeeded, False or exception otherwise
    """
    mandatoryArgs = set(["GPUMemoryMB", "CUDARuntime", "CUDACapabilities"])
    optionalArgs = set(["GPUName", "CUDADriverVersion", "CUDARuntimeVersion"])
    try:
        data = json.loads(candidate)
    except Exception:
        raise AssertionError("GPUParams is not a valid JSON object")
        # once python2 code is deprecated, this is the way to raise only the last exception
        # raise AssertionError("GPUParams is not a valid JSON object") from None
    if data is None:
        return True
    if not isinstance(data, dict):
        raise AssertionError("GPUParams is not a dictionary encoded as JSON object")

    paramSet = set(viewkeys(data))
    # is every mandatory argument also in the provided args?
    if not mandatoryArgs <= paramSet:
        msg = "GPUParams does not contain all the mandatory arguments. "
        msg +="Mandatory args: {}, while args provided are: {}".format(mandatoryArgs, paramSet)
        raise AssertionError(msg)
    # are there unknown arguments in the data provided?
    unknownArgs = paramSet - mandatoryArgs - optionalArgs
    if unknownArgs:
        msg = "GPUParams contains arguments that are not supported. Args provided: {}, ".format(paramSet)
        msg +="while mandatory args are: {} and optional args are: {}".format(mandatoryArgs, optionalArgs)
        raise AssertionError(msg)
    return _gpuInternalParameters(data)


CUDA_VERSION_REGEX = {"re": r"^\d+\.\d+(\.\d+)?$", "maxLength": 100}
def _gpuInternalParameters(candidate):
    """
    NOTE: this function is supposed to be called only from gpuParameters, which already
    does the high level validation.
    List of **required** parameters is:
      * `GPUMemoryMB`: integer with the amount of memory, in Megabytes (MB). Validate as `> 0`. E.g.: 8000
      * `CUDACapabilities`: a list of short strings (<= 100 chars). Validation should ensure at least one item
        in the list and matching this regex: `r"^\d+.\d$"`. E.g.: ["7.5", "8.0"]
      * `CUDARuntime`: a short string (<=100 chars) with the runtime version.
        Validated against this regex: `r"^\d+.\d+$"`. E.g.: "11.2"
    List of **optional** parameters is:
      * `GPUName`: a string with the GPU name. Validate against `<= 100 chars`. E.g. "Tesla T4", "Quadro RTX 6000";
      * `CUDADriverVersion`: a string with the CUDA driver version.
        Validated against this regex: `r"^\d+.\d+\d+$"`E.g. "460.32.03"
      * `CUDARuntimeVersion`: a string with the CUDA runtime version.
        Validated against this regex: `r"^\d+.\d+\d+$"`E.g. "11.2.152"

    This function validates all the internal key/value pairs provided for the GPUParams
    argument, mostly against their own regular expressions.
    :param candidate: the JSON object already decoded (thus, str or dict)
    :return: True if validation succeeded, False or exception otherwise
    """
    # Generic regular expression for CUDA runtime/driver version
    # It matches either something like "11.2", or "11.2.231"
    # GPUMemoryMB validation
    if not isinstance(candidate["GPUMemoryMB"], int) or not candidate["GPUMemoryMB"] > 0:
        raise AssertionError("Mandatory GPUParams.GPUMemoryMB must be an integer and greater than 0")
    # CUDACapabilities validation
    if not isinstance(candidate["CUDACapabilities"], (list, set)) or not candidate["CUDACapabilities"]:
        raise AssertionError("Mandatory GPUParams.CUDACapabilities must be a non-empty list")
    for cudaCapabItem in candidate["CUDACapabilities"]:
        if not isinstance(cudaCapabItem, (str, bytes)):
            raise AssertionError("Mandatory GPUParams.CUDACapabilities must be a list of strings")
        check(CUDA_VERSION_REGEX["re"], cudaCapabItem, CUDA_VERSION_REGEX["maxLength"])
    # CUDARuntime validation
    if not isinstance(candidate["CUDARuntime"], (str, bytes)) or\
            not check(CUDA_VERSION_REGEX["re"], candidate["CUDARuntime"], CUDA_VERSION_REGEX["maxLength"]):
        raise AssertionError("Mandatory GPUParams.CUDARuntime must be a string and shorter than 100 chars")

    ### And now, validate the optional arguments
    # GPUName validation
    if "GPUName" in candidate:
        if not isinstance(candidate["GPUName"], (str, bytes)):
            raise AssertionError("Optional GPUParams.GPUName must be a string")
        check(r".*", candidate["GPUName"], 100)
    # CUDADriverVersion validation
    if "CUDADriverVersion" in candidate:
        if not isinstance(candidate["CUDADriverVersion"], (str, bytes)):
            raise AssertionError("Optional GPUParams.CUDADriverVersion must be a string")
        check(CUDA_VERSION_REGEX["re"], candidate["CUDADriverVersion"], CUDA_VERSION_REGEX["maxLength"])
    # CUDARuntimeVersion validation
    if "CUDARuntimeVersion" in candidate:
        if not isinstance(candidate["CUDARuntimeVersion"], (str, bytes)):
            raise AssertionError("Optional GPUParams.CUDARuntimeVersion must be a string")
        check(CUDA_VERSION_REGEX["re"], candidate["CUDARuntimeVersion"], CUDA_VERSION_REGEX["maxLength"])
    return True


def getStringsBetween(start, end, source):
    """
    get the string between start string and end string for given source string
    source string is one line string (no new line charactor in it)
    """
    regex = r"\%s(.*?)\%s" % (start, end)
    result = re.match(regex, source)
    if result:
        return result.group(1).strip()
    else:
        return None


def getIterMatchObjectOnRegexp(filePath, regexp):
    with io.open(filePath, 'r', encoding='utf8', errors='ignore') as f:
        for m in re.finditer(regexp, f.read()):
            yield m



#!/usr/bin/env python
"""
_MarkNewFinishedSubscriptions_

Oracle implementation of Subscription.MarkNewFinishedSubscriptions
"""

from WMCore.WMBS.MySQL.Subscriptions.MarkNewFinishedSubscriptions \
        import MarkNewFinishedSubscriptions as MySQLMarkNewFinishedSubscriptions

class MarkNewFinishedSubscriptions(MySQLMarkNewFinishedSubscriptions):
    pass



#!/usr/bin/env python
# pylint: disable=W0613

"""
Creates jobs for new subscriptions
Handler implementation for polling

"""
from __future__ import print_function

import logging
import threading

from WMComponent.JobCreator.JobCreatorPoller import JobCreatorPoller
# harness class that encapsulates the basic component logic.
from WMCore.Agent.Harness import Harness


class JobCreator(Harness):
    """
    Creates jobs for new subscriptions
    Handler implementation for polling

    """

    def __init__(self, config):
        # call the base class
        Harness.__init__(self, config)
        self.pollTime = 1

        # myThread = threading.currentThread()
        # myThread.database = os.getenv("DATABASE")

        print("JobCreator.__init__")

    def preInitialization(self):
        """
        Step that actually adds the worker thread properly

        """
        print("JobCreator.preInitialization")

        # Add event loop to worker manager
        myThread = threading.currentThread()

        pollInterval = self.config.JobCreator.pollInterval
        logging.info("Setting poll interval to %s seconds", pollInterval)
        myThread.workerThreadManager.addWorker(JobCreatorPoller(self.config), pollInterval)

        return



#!/usr/bin/env python
"""
_DBSWriterObjects_

Functions to instantiate and return DBS Objects and insert them
into DBS if required

"""
from __future__ import print_function

from builtins import int

import logging

from DBSAPI.dbsException import *
from DBSAPI.dbsApiException import *
from DBSAPI.dbsPrimaryDataset import DbsPrimaryDataset
from DBSAPI.dbsAlgorithm import DbsAlgorithm
from DBSAPI.dbsQueryableParameterSet import DbsQueryableParameterSet
from DBSAPI.dbsProcessedDataset import DbsProcessedDataset
from DBSAPI.dbsFile import DbsFile
from DBSAPI.dbsFileBlock import DbsFileBlock
from DBSAPI.dbsStorageElement import DbsStorageElement
from DBSAPI.dbsRun import DbsRun
from DBSAPI.dbsLumiSection import DbsLumiSection

def makeTierList(dataTier):
    """
    _makeTierList_

    Standard tool to split data tiers if they contain - chars
    *** Do not use outside of this module ***

    """
    tierList = dataTier.split("-")
    return tierList

def createPrimaryDataset(datasetInfo, apiRef = None):
    """
    _createPrimaryDataset_

    Create and return a Primary Dataset object.
    If apiRef is not None, it is used to insert the dataset into the
    DBS

    """
    if 'PrimaryDatasetType' in datasetInfo:
        PrimaryDatasetType = datasetInfo['PrimaryDatasetType']
    else:
        PrimaryDatasetType = 'mc'

    logging.debug("Inserting PrimaryDataset %s with Type %s", datasetInfo["PrimaryDataset"], PrimaryDatasetType)
    primary = DbsPrimaryDataset(Name = datasetInfo["PrimaryDataset"], Type=PrimaryDatasetType)

    if apiRef != None:
        apiRef.insertPrimaryDataset(primary)
    return primary


def createAlgorithm(datasetInfo, configMetadata = None, apiRef = None):
    """
    _createAlgorithm_

    Create an algorithm assuming that datasetInfo is a
    ProdCommon.MCPayloads.DatasetInfo like dictionary

    """

    exeName = datasetInfo['ApplicationName']
    appVersion = datasetInfo['ApplicationVersion']
    appFamily = datasetInfo["ApplicationFamily"]

    #
    # HACK:  Problem with large PSets (is this still relevant ?)
    #
    # Repacker jobs have no PSetContent/PSetHash
    #
    psetContent = datasetInfo.get('PSetContent',None)
    if psetContent == None:
        psetContent = "PSET_CONTENT_NOT_AVAILABLE"
    psetHash = datasetInfo.get('PSetHash',None)
    if psetHash == None:
        psetHash = "NO_PSET_HASH"
    else:
        if psetHash.find(";"):
            # no need for fake hash in new schema
            psetHash = psetHash.split(";")[0]
            psetHash = psetHash.replace("hash=", "")

    ## No more hacks
    #msg = ">>>>>>>>>>>>>>>>>>>>>>>>>>>>\n"
    #msg += "TEST HACK USED FOR PSetContent\n"
    #msg += ">>>>>>>>>>>>>>>>>>>>>>>>>>>>"
    #logging.warning(msg)
    #print msg
    #psetContent = "This is not a PSet"

    #
    # HACK: 100 char limit on cfg file name
    if configMetadata != None:
        cfgName = configMetadata['name']
        if len(cfgName) > 100:
            msg = ">>>>>>>>>>>>>>>>>>>>>>>>>>>>\n"
            msg += "TEST HACK USED FOR Config File Name"
            msg += ">>>>>>>>>>>>>>>>>>>>>>>>>>>>"
            logging.warning(msg)
            print(msg)
            configMetadata['name'] = cfgName[-99]

        psetInstance = DbsQueryableParameterSet(
            Hash = psetHash,
            Name = configMetadata['name'],
            Version = configMetadata['version'],
            Type = configMetadata['Type'],
            Annotation = configMetadata['annotation'],
            Content = psetContent,
            )


        algorithmInstance = DbsAlgorithm(
            ExecutableName = exeName,
            ApplicationVersion = appVersion,
            ApplicationFamily = appFamily,
            ParameterSetID = psetInstance
            )
    else:
        psetInstance = DbsQueryableParameterSet(
                    Hash = psetHash)
        algorithmInstance = DbsAlgorithm(
            ExecutableName = exeName,
            ApplicationVersion = appVersion,
            ApplicationFamily = appFamily,
            ParameterSetID = psetInstance
            )

    if apiRef != None:
        apiRef.insertAlgorithm(algorithmInstance)
    return algorithmInstance

def createAlgorithmForInsert(datasetInfo):
    """
    _createPartialAlgorithm_

    Create an Algorithm instance that uses the minimal info needed
    to insert a file

    """
    exeName = datasetInfo['ApplicationName']
    appVersion = datasetInfo['ApplicationVersion']
    appFamily = datasetInfo["ApplicationFamily"]

    #
    # Repacker jobs have no PsetContent/PSetHash
    #
    psetContent = datasetInfo.get('PSetContent',None)
    if psetContent == None:
        psetContent = "PSET_CONTENT_NOT_AVAILABLE"
    psetHash = datasetInfo.get('PSetHash',None)
    if psetHash == None:
        psetHash = "NO_PSET_HASH"
    else:
        if psetHash.find(";"):
            # no need for fake hash in new schema
            psetHash = psetHash.split(";")[0]
            psetHash = psetHash.replace("hash=", "")

    psetInstance = DbsQueryableParameterSet(
        Hash = psetHash)
    algorithmInstance = DbsAlgorithm(
        ExecutableName = exeName,
        ApplicationVersion = appVersion,
        ApplicationFamily = appFamily,
        ParameterSetID = psetInstance
        )
    return algorithmInstance

def createMergeAlgorithm(datasetInfo, apiRef = None):
    """
    _createMergeAlgorithm_

    Create a DbsAlgorithm for a merge dataset

    """
    exeName = datasetInfo['ApplicationName']
    version = datasetInfo['ApplicationVersion']
    family = datasetInfo.get('ApplicationFamily', None)
    if (family == None) or not (family) :
        family = datasetInfo['OutputModuleName']


    mergeAlgo = DbsAlgorithm (
        ExecutableName = exeName,
        ApplicationVersion = version,
        ApplicationFamily = family,
        )

    if apiRef != None:
        apiRef.insertAlgorithm(mergeAlgo)
    return mergeAlgo




def createProcessedDataset(primaryDataset, algorithm, datasetInfo,
                           apiRef = None):
    """
    _createProcessedDataset_


    """

    physicsGroup = datasetInfo.get("PhysicsGroup", "NoGroup")
    status = datasetInfo.get("Status", "VALID")
    dataTier = datasetInfo['DataTier']
    globalTag = datasetInfo.get('Conditions', None)
    if globalTag is None: globalTag = ''

    parents = []
    inputDataset = datasetInfo.get('ParentDataset', None)
    if inputDataset != None:
        parents.append(inputDataset)

    tierList = makeTierList(datasetInfo['DataTier'])

    name = datasetInfo['ProcessedDataset']
    algolist=[]
    if algorithm not in ('', None):
        algolist=list(algorithm)

    processedDataset = DbsProcessedDataset (
        PrimaryDataset = primaryDataset,
        AlgoList=algolist,
        Name = name,
        TierList = tierList,
        ParentList = parents,
        PhysicsGroup = physicsGroup,
        Status = status,
        GlobalTag = globalTag,
        )

    if apiRef != None:
        apiRef.insertProcessedDataset(processedDataset)
    #
    logging.debug("PrimaryDataset: %s ProcessedDataset: %s DataTierList: %s  requested by PhysicsGroup: %s ", primaryDataset['Name'], name, tierList, physicsGroup)
    return processedDataset




def createDBSFiles(fjrFileInfo, jobType = None, apiRef = None):
    """
    _createDBSFiles_

    Create a list of DBS File instances from the file details contained
    in a FwkJobRep.FileInfo instance describing an output file
    Does not insert files, returns as list of DbsFile objects
    Does insert runs and lumisections if DBS API reference is passed

    """
    results = []
    inputLFNs = [ x['LFN'] for x in fjrFileInfo.inputFiles]
    checksum = fjrFileInfo.checksums['cksum']
    adler32sum = fjrFileInfo.checksums.get('adler32', '')

    nEvents = int(fjrFileInfo['TotalEvents'])

    if len(fjrFileInfo.dataset)<=0:
        logging.error("No dataset info found in FWJobReport!")
        return results

    #  //
    # // Set FileType
    #//
    if 'FileType' in fjrFileInfo:
        fileType = fjrFileInfo['FileType']
    else:
        fileType = 'EDM'

    #
    # FIXME: at this point I should use the mc or data event type from
    #        the jobreport. Until this is supported by the framework,
    #        we use the workaround that mc job reports have an empty
    #        lumisections list (stripped in DBSInterface)
    #
    lumiList = []
    if ( len(fjrFileInfo.getLumiSections()) > 0 ):

        #
        # insert runs (for data files from detector)
        #
        if ( apiRef != None ):

            for runinfo in fjrFileInfo.runs:

                run = DbsRun(
                    RunNumber = int(runinfo),
                    NumberOfEvents = 0,
                    NumberOfLumiSections = 0,
                    TotalLuminosity = 0,
                    StoreNumber = 0,
                    StartOfRun = 0,
                    EndOfRun = 0,
                    )

                apiRef.insertRun(run)

        #
        # insert lumisections (for data files from detector)
        # associate files with lumisections (for all data files)
        #
        for lumiinfo in fjrFileInfo.getLumiSections():

            lumi = DbsLumiSection(
                LumiSectionNumber = int(lumiinfo['LumiSectionNumber']),
                StartEventNumber = 0,
                EndEventNumber = 0,
                LumiStartTime = 0,
                LumiEndTime = 0,
                RunNumber = int(lumiinfo['RunNumber']),
                )

            # Isnt needed, causes monster slowdown
            #if ( apiRef != None ):
            #    apiRef.insertLumiSection(lumi)

            lumiList.append(lumi)

            logging.debug("Lumi associated to file is: %s" % ([x for x in lumiList]))

    #  //
    # // Dataset info related to files and creation of DbsFile object
    #//
    for dataset in fjrFileInfo.dataset:

        primary = createPrimaryDataset(dataset)
        if jobType == "Merge":
            algo = createMergeAlgorithm(dataset)
        else:
            algo = createAlgorithmForInsert(dataset)

        processed = createProcessedDataset(primary, algo, dataset)

        dbsFileInstance = DbsFile(
            Checksum = checksum,
            Adler32 = adler32sum,
            NumberOfEvents = nEvents,
            LogicalFileName = fjrFileInfo['LFN'],
            FileSize = int(fjrFileInfo['Size']),
            Status = "VALID",
            ValidationStatus = 'VALID',
            FileType = fileType,
            Dataset = processed,
            TierList = makeTierList(dataset['DataTier']),
            AlgoList = [algo],
            LumiList = lumiList,
            ParentList = inputLFNs,
            BranchList = fjrFileInfo.branches,
            )

        results.append(dbsFileInstance)
    return results


def createDBSStorageElement(pnn):
    """
    _createDBSStorageElement_

    """
    return DbsStorageElement(Name = pnn)


def createDBSFileBlock(blockName):
    """
    _createDBSFileBlock_

    return a DbsFileBlock object with the block name provided

    NOTE: This method DOES NOT create a new block in DBS

    """
    return DbsFileBlock(Name=blockName)



#!/usr/bin/env python
"""
_IDFromFilesetWorkflow_

MySQL implementation of Subscription.IDFromFilesetWorkflow
"""




from WMCore.Database.DBFormatter import DBFormatter

class IDFromFilesetWorkflow(DBFormatter):
    sql = """SELECT id FROM wmbs_subscription
             WHERE fileset = (SELECT id FROM wmbs_fileset WHERE name = :fileset)
             AND workflow = (SELECT id FROM wmbs_workflow WHERE name = :workflow)
             """
    def format(self, result):
        result = DBFormatter.format(self, result)

        if len(result) > 0:
            return result[0][0]
        else:
            return False

    def execute(self, workflow = None, fileset = None, conn = None,
                transaction = False):
        result = self.dbi.processData(self.sql, {"workflow": workflow, "fileset": fileset},
                                      conn = conn, transaction = transaction)
        return self.format(result)



#!/usr/bin/env python
"""
_GetCouchID_

MySQL implementation of Jobs.GetCouchID
"""




from WMCore.Database.DBFormatter import DBFormatter

class GetCouchID(DBFormatter):
    """
    _GetCouchID_

    Given a job ID retrieve the couch document ID.
    """
    sql     = "SELECT couch_record FROM wmbs_job WHERE id = :jobid"

    bulkSQL = """SELECT couch_record AS couch_record, id AS jobid FROM wmbs_job WHERE id = :jobid"""

    def format(self, results):
        """
        _format_

        Return the couch document ID or None if one has not been set.
        """
        result = DBFormatter.format(self, results)

        if len(result) == 0:
            return None

        return result[0][0]

    def execute(self, jobID, conn = None, transaction = False):
        """
        _execute_

        Execute the SQL for the given job ID and then format and return
        the result.
        """
        if isinstance(jobID, list):
            if len(jobID) == 0:
                return {}
            binds = []
            for entry in jobID:
                binds.append({'jobid': entry})

            result = self.dbi.processData(self.bulkSQL, binds, conn = conn,
                                          transaction = transaction)

            return self.formatDict(result)


        result = self.dbi.processData(self.sql, {"jobid": jobID}, conn = conn,
                                      transaction = transaction)

        return self.format(result)



#!/usr/bin/env python
"""
_GetParentAndGrandParentInfo_

Oracle implementation of Files.GetParentAndGrandParentInfo
"""
from __future__ import division

from WMCore.WMBS.MySQL.Files.GetParentAndGrandParentInfo import \
    GetParentAndGrandParentInfo as GetParentAndGrandParentInfoMySQL


class GetParentAndGrandParentInfo(GetParentAndGrandParentInfoMySQL):
    pass



#!/usr/bin/env python
"""
_TestImpl_

Couple of test implementations for unittests

"""

import os
import os.path
import shutil


from WMCore.Storage.StageOutImplV2 import StageOutImplV2
from WMCore.Storage.StageOutError import StageOutFailure


class TestLocalCopyImpl(StageOutImplV2):
    """
    _LocalCopyImp_

    Test plugin that copies to a local directory

    """

    def createOutputDirectory(self, targetPFN):
        """
        I guess this masks a directory?

        """

        dirName = os.path.dirname(targetPFN)

        if not os.path.isdir(dirName):
            os.makedirs(dirName)

        return

    def doTransfer(self, fromPfn, toPfn, stageOut, pnn, command, options, protocol, checksum  ):
        self.createOutputDirectory( toPfn )
        shutil.copy(fromPfn, toPfn)
        if os.path.getsize(fromPfn) != os.path.getsize(toPfn):
            raise StageOutFailure("Invalid file size")
        return toPfn


    def doDelete(self, pfn, pnn, command, options, protocol  ):
        os.unlink(pfn)



#!/usr/bin/env python
# encoding: utf-8
"""
CouchWorkQueueElement.py

Created by Dave Evans on 2010-10-12.
Copyright (c) 2010 Fermilab. All rights reserved.
"""

from future.utils import viewitems

import unittest
import time
import logging

from WMCore.Database.CMSCouch import Document
from WMCore.WorkQueue.DataStructs.WorkQueueElement import WorkQueueElement



class CouchWorkQueueElement(WorkQueueElement):
    """
    _CouchWorkQueueElement_

    """
    def __init__(self, couchDB, id = None, elementParams = None):
        elementParams = elementParams or {}
        WorkQueueElement.__init__(self, **elementParams)
        if id:
            self._id = id
        self._document = Document(id = id)
        self._couch = couchDB

    rev = property(
        lambda x: str(x._document[u'_rev']) if u'_rev' in x._document else x._document.__getitem__('_rev'),
        lambda x, newid: x._document.__setitem__('_rev', newid))
    timestamp = property(
        lambda x: str(x._document[u'timestamp']) if u'timestamp' in x._document else x._document.__getitem__('timestamp')
        )
    updatetime = property(
        lambda x: str(x._document[u'updatetime']) if u'updatetime' in x._document else 0
        )


    @classmethod
    def fromDocument(cls, couchDB, doc):
        """Create element from couch document"""
        elementParams = doc.pop('WMCore.WorkQueue.DataStructs.WorkQueueElement.WorkQueueElement')
        elementParams["CreationTime"] = doc.pop('timestamp')
        element = CouchWorkQueueElement(couchDB = couchDB,
                                        id = doc['_id'],
                                        elementParams = elementParams)
        element._document['_rev'] = doc.pop('_rev')
        element._document['timestamp'] = elementParams["CreationTime"]
        element._document['updatetime'] = doc.pop('updatetime')
        return element

    def save(self):
        """
        _save
        """
        self.populateDocument()
        self._couch.queue(self._document)

    def load(self):
        """
        _load_

        Load the document representing this WQE
        """
        document = self._couch.document(self._document['_id'])
        self.update(document.pop('WMCore.WorkQueue.DataStructs.WorkQueueElement.WorkQueueElement'))
        self._document['_rev'] = document.pop('_rev')
        self._document['timestamp'] = document.pop('timestamp', None)
        self._document['updatetime'] = document.pop('updatetime', None)
        return self

    def delete(self):
        """Delete element"""
        self.populateDocument()
        self._document.delete()
        self._couch.queue(self._document)

    def populateDocument(self):
        """Certain attributed shouldn't be stored"""
        self._document.update(self.__to_json__(None))
        now = time.time()
        self._document['updatetime'] = now
        self._document.setdefault('timestamp', now)
        if not self._document.get('_id') and self.id:
            self._document['_id'] = self.id
        attrs = ['WMSpec', 'Task']
        for attr in attrs:
            self._document['WMCore.WorkQueue.DataStructs.WorkQueueElement.WorkQueueElement'].pop(attr, None)



def fixElementConflicts(*elements):
    """Take conflicting elements, fix conflicts and return
    First element returned will contain the merged results,
    the others will be losing revisions
    """
    ordered_states = ['Available', 'Negotiating', 'Acquired', 'Running',
                      'Done', 'Failed', 'CancelRequested', 'Canceled']
    allowed_keys = ['Status', 'EventsWritten', 'FilesProcessed',
                    'PercentComplete', 'PercentSuccess', 'Inputs', 'NumOfFilesAdded',
                    'SubscriptionId', 'Priority', 'SiteWhitelist', 'SiteBlacklist']
    merged_value = None
    updated = set()
    for ele in elements:
        if not merged_value: # 1st will contain merged result and become winner
            merged_value = ele
            continue

        # print differences
        from WMCore.Algorithms.MiscAlgos import dict_diff
        logging.info("Conflict between %s revs %s & %s: %s",
                         ele.id, merged_value.rev, ele.rev,
                         "; ".join("%s=%s" % (x,y) for x,y in viewitems(dict_diff(merged_value, ele)))
                    )
        for key in merged_value:
            if merged_value[key] == ele.get(key):
                continue
            # we need to merge: Take elements from both that seem most advanced, e.g. status & progress stats
            if key not in allowed_keys:
                msg = 'Unable to merge conflicting element %s: field "%s" value 1 "%s" value2 "%s"'
                raise RuntimeError(msg % (ele.id, key, merged_value.get(key), ele.get(key)))
            elif key == 'Status':
                if ordered_states.index(ele[key]) > ordered_states.index(merged_value[key]):
                    merged_value[key] = ele[key]
            elif key == 'Inputs':
                for item in merged_value[key]:
                    # take larger locations list
                    if merged_value[key][item] < ele[key].get(item, []):
                        merged_value[key][item] = ele[key][item]
            elif ele[key] is not None and ele[key] > merged_value[key]:
                merged_value[key] = ele[key]
            updated.add(key)
        # once losing element has been merged - queue for deletion
        ele._document.delete()

    msg = 'Resolving conflict for wf "%s", id "%s": Remove rev(s): %s: Updates: (%s)'
    logging.info(msg, str(merged_value['RequestName']),
                 str(merged_value.id),
                 ", ".join([x._document['_rev'] for x in elements[1:]]),
                 "; ".join("%s=%s" % (x, merged_value[x]) for x in updated)
                 )
    return elements



#!/usr/bin/env python
"""
_InsertOutput_

Oracle implementation of Workflow.InsertOutput
"""

from WMCore.WMBS.MySQL.Workflow.InsertOutput import InsertOutput as InsertOutputMySQL

class InsertOutput(InsertOutputMySQL):
    pass



#!/usr/bin/env python
"""
_LoadComplete_

MySQL implementation for loading bl_runjob records for complete jobs
"""
import logging

from WMCore.Database.DBFormatter import DBFormatter

class LoadComplete(DBFormatter):
    """
    _LoadComplete_

    Load all bl_runjob that are complete, where the
    corresponding wmbs_job entry is still in 'executing'
    and where the retry_count match between bl_runjob
    and wmbs_job. The later makes sure we only look at
    the latest bl_runjob instance for a given wmbs_job.

    The two query implementations give identical results,
    in a congested database (few wmbs_job in executing
    and almost all bl_runjob with status 0) the second
    one is likely faster (not tested).

    """

    sql = """SELECT bl_runjob.wmbs_id AS jobid,
                    bl_runjob.grid_id AS gridid,
                    bl_runjob.bulk_id AS bulkid,
                    bl_status.name AS status,
                    bl_runjob.retry_count AS retry_count,
                    bl_runjob.id AS id,
                    bl_runjob.status_time AS status_time,
                    wmbs_job.name AS name,
                    wmbs_location.ce_name AS location,
                    wmbs_users.cert_dn AS userdn,
                    wmbs_users.group_name AS usergroup, wmbs_users.role_name AS userrole
             FROM wmbs_job
             INNER JOIN bl_runjob ON
               bl_runjob.wmbs_id = wmbs_job.id AND
               bl_runjob.retry_count = wmbs_job.retry_count AND
               bl_runjob.status = 0
             INNER JOIN bl_status ON
               bl_status.id = bl_runjob.sched_status
             LEFT OUTER JOIN wmbs_users ON wmbs_users.id = bl_runjob.user_id
             INNER JOIN wmbs_location ON wmbs_location.id = bl_runjob.location
             WHERE wmbs_job.state = (SELECT id FROM wmbs_job_state WHERE name = 'executing')
             """

    def execute(self, conn = None, transaction = False):
        """
        _execute_

        """
        result = self.dbi.processData(self.sql, binds = {}, conn = conn,
                                      transaction = transaction)

        return self.formatDict(result)



from __future__ import print_function
from glob import iglob
import shutil
import os

WMSTATS_SCRIPT_DIR = "../src/couchapps/WMStats/_attachments/js"
MINIFIED_DIR = "../src/couchapps/WMStats/_attachments/js/minified"

globalLib = ["WMStats.Globals.js",
             "WMStats.Utils.js",
             "WMStats.Couch.js",
             "WMStats.Ajax.js"]

dataStrunct = ["DataStruct/WMStats._StructBase.js",
               "DataStruct/WMStats.GenericRequests.js",
               "DataStruct/WMStats.Tasks.js",
               "DataStruct/WMStats.Agents.js",
               "DataStruct/WMStats.Sites.js",
               "DataStruct/WMStats.JobSummary.js",
               "DataStruct/WMStats.Campaigns.js",
               "DataStruct/WMStats.Alerts.js",
               "DataStruct/WMStats.SiteSummary.js",
               "DataStruct/WMStats.JobDetails.js",
               "DataStruct/WMStats.WorkloadSummary.js",
               "DataStruct/WMStats.History.js",
               "DataStruct/WMStats.LogDBData.js",
               "DataStruct/WMStats.LogMessage.js"]

# tier1 specific library
t1_dataStruct = ["DataStruct/T1/WMStats.RequestSummary.js",
                "DataStruct/T1/WMStats.CampaignSummary.js",
                "DataStruct/T1/WMStats.CMSSWSummary.js",
                "DataStruct/T1/WMStats.AgentRequestSummary.js",
                "DataStruct/T1/WMStats.ReqMgrRequest.js"]


# tier0 specific library
t0_dataStruct = ["DataStruct/T0/WMStats.RequestSummary.js",
                "DataStruct/T0/WMStats.RunSummary.js"]

viewModel = ["ViewModels/WMStats.ViewModel.js"]

views = ["Views/WMStats.CategoryMap.js",
         "Views/WMStats.View.IndexHTML.js",
         "Views/Controls/WMStats.CommonControls.js",
         "Views/Tables/WMStats.Table.js",
         "Views/Tables/WMStats.JobSummaryTable.js",
         "Views/Tables/WMStats.WorkloadSummaryTable.js",
         "Views/Tables/WMStats.TableController.js",
         "Views/HTMLList/WMStats.JobDetailList.js",
         "Views/HTMLList/WMStats.AgentDetailList.js",
         "Views/HTMLList/WMStats.CategoryTitle.js",
         "Views/HTMLList/WMStats.RequestTitle.js",
         "Views/HTMLList/WMStats.RequestLogList.js",
         "Views/Graphs/WMStats.SiteHistoryGraph.js"
         ]

# tier1 specific library
t1_views = ["Views/Controls/T1/WMStats.Controls.js",
            "Views/Tables/T1/WMStats.ActiveRequestTable.js",
            "Views/Tables/T1/WMStats.ActiveRequestTableWithJob.js",
            "Views/Tables/T1/WMStats.TaskSummaryTable.js",
            "Views/Tables/T1/WMStats.CampaignSummaryTable.js",
            "Views/Tables/T1/WMStats.CMSSWSummaryTable.js",
            "Views/Tables/T1/WMStats.SiteSummaryTable.js",
            "Views/Tables/T1/WMStats.AgentRequestSummaryTable.js",
            "Views/Tables/T1/addCategoryMap.js",
            "Views/HTMLList/T1/WMStats.RequestDetailList.js",
            "Views/HTMLList/T1/WMStats.RequestAlertGUI.js",
            "Views/HTMLList/T1/WMStats.CategoryDetailList.js",
            "Views/HTMLList/T1/WMStats.RequestSummaryList.js",
            "Views/HTMLList/T1/WMStats.RequestDataList.js",
            "Views/HTMLList/T1/WMStats.ResubmissionList.js"]

# tier1 specific library
t0_views = ["Views/Controls/T0/WMStats.Controls.js",
            "Views/Tables/T0/WMStats.ActiveRequestTable.js",
            "Views/Tables/T0/WMStats.TaskSummaryTable.js",
            "Views/Tables/T0/WMStats.RunSummaryTable.js",
            "Views/Tables/T0/addCategoryMap.js",
            "Views/HTMLList/T0/WMStats.RequestDetailList.js",
            "Views/HTMLList/T0/WMStats.RequestAlertGUI.js",
            "Views/HTMLList/T0/WMStats.CategoryDetailList.js",
            "Views/HTMLList/T0/WMStats.RequestSummaryList.js",
            "Views/HTMLList/T0/WMStats.RequestDataList.js"]

models =["Models/WMStats._ModelBase.js",
         "Models/WMStats._AjaxModelBase.js",
         "Models/WMStats._RequestModelBase.js",
         "Models/WMStats.JobSummaryModel.js",
         "Models/WMStats.JobDetailModel.js",
         "Models/WMStats.AgentModel.js",
         "Models/WMStats.WorkloadSummaryModel.js",
         "Models/WMStats.HistoryModel.js"]

t1_models =["Models/T1/WMStats.ActiveRequestModel.js",
            "Models/T1/WMStats.ReqMgrRequestModel.js",
            "Models/T1/WMStats.RequestSearchModel.js",
            "Models/T1/WMStats.RequestLogDetailModel.js",
            "Models/T1/WMStats.RequestLogModel.js"]

t0_models =["Models/T0/WMStats.ActiveRequestModel.js",
            "Models/T0/WMStats.RequestModel.js",
            "Models/T0/WMStats.RequestSearchModel.js",
            "Models/T0/WMStats.RequestLogDetailModel.js",
            "Models/T0/WMStats.RequestLogModel.js"]


controller = ["Controller/WMStats.GenericController.js",
             "Controller/WMStats.ActiveRequestController.js",
             "Controller/WMStats.WorkloadSummaryController.js"]


def concatenateFiles(filelist, fileName):
    destPath = os.path.join(MINIFIED_DIR, fileName)
    destination = open(destPath, 'wb')
    for filename in filelist:
        filePath = os.path.join(WMSTATS_SCRIPT_DIR, filename)
        shutil.copyfileobj(open(filePath, 'rb'), destination)
    destination.close()
    try:
        from subprocess import call
        print(destPath)
        call(["uglifyjs", destPath, "-o", destPath])
    except Exception as ex:
        print("%s" % str(ex))
        pass

if __name__ == "__main__":

    initLoadDest = 'global.min.js'
    t1Dest = 'import-all-t1.min.js'
    t0Dest = 'import-all-t0.min.js'

    concatenateFiles(globalLib, initLoadDest)

    t1_lib = []
    t1_lib.extend(dataStrunct)
    t1_lib.extend(t1_dataStruct)
    t1_lib.extend(viewModel)
    t1_lib.extend(views)
    t1_lib.extend(t1_views)
    t1_lib.extend(models)
    t1_lib.extend(t1_models)
    t1_lib.extend(controller)
    concatenateFiles(t1_lib, t1Dest)

    t0_lib = []
    t0_lib.extend(dataStrunct)
    t0_lib.extend(t0_dataStruct)
    t0_lib.extend(viewModel)
    t0_lib.extend(views)
    t0_lib.extend(t0_views)
    t0_lib.extend(models)
    t0_lib.extend(t0_models)
    t0_lib.extend(controller)
    concatenateFiles(t0_lib, t0Dest)





#!/usr/bin/env python

from __future__ import division, print_function

from WMCore.Database.DBFormatter import DBFormatter


class GetSubsWithoutJobGroup(DBFormatter):
    """
    _GetSubsWithoutJobGroup_

    Finds whether there are unfinished subscriptions for Production and
    Processing task types where JobCreator hasn't yet created any jobs
    nor a jobgroup associated to it.
    """

    sql = """SELECT wmbs_subscription.id, wmbs_workflow.task FROM wmbs_subscription
               INNER JOIN wmbs_sub_types ON wmbs_sub_types.id = wmbs_subscription.subtype
               INNER JOIN wmbs_workflow ON wmbs_workflow.id = wmbs_subscription.workflow
               WHERE wmbs_subscription.finished=0 AND
                 wmbs_sub_types.name IN ('Production','Processing') AND
                 NOT EXISTS (SELECT * FROM wmbs_jobgroup
                   WHERE wmbs_jobgroup.subscription = wmbs_subscription.id)
          """

    def format(self, result):
        """
        Have to filter task names that contain only two slashes '/',
        such that we can declare those tasks as top level task.
        :param result: 
        :return: a list of subscriptions id
        """
        results = DBFormatter.format(self, result)

        subIDs = []
        for row in results:
            if len(row[1].split('/')) <= 3:  # remember, first item is empty
                subIDs.append(row[0])

        return subIDs

    def execute(self, conn=None, transaction=False):
        result = self.dbi.processData(self.sql, conn=conn, transaction=transaction)
        return self.format(result)



#!/usr/bin/env python
"""
_ListSites_

MySQL implementation of Locations.ListSites
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

from future.utils import listvalues
import logging

class ListSites(DBFormatter):
    sql = "SELECT site_name FROM wmbs_location"


    def format(self, results):

        if len(results) == 0:
            return False
        else:
            format = []
            for i in results[0].fetchall():
                format.append(listvalues(i)[0])
            return format


    def execute(self, conn = None, transaction = False):

        results = self.dbi.processData(self.sql, {}, conn = conn, transaction = transaction)
        return self.format(results)



#!/usr/bin/env python
"""
_Delete_

MySQL implementation of Masks.Delete
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class Delete(DBFormatter):
    sql = "delete from wmbs_job_mask where job = :jobid"

    def execute(self, jobid, conn = None, transaction = False):
        binds = self.getBinds(jobid = jobid)
        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)
        return



#!/usr/bin/env python
"""
_AddBulkParentage_

MySQL implementation of Files.AddBulkParentage
"""




from WMCore.Database.DBFormatter import DBFormatter

class AddBulkParentage(DBFormatter):
    sql = "INSERT INTO wmbs_file_parent (child, parent) VALUES (:child, :parent)"

    def execute(self, fileParentage, conn = None, transaction = False):
        self.dbi.processData(self.sql, fileParentage, conn = conn,
                             transaction = transaction)
        return



#!/usr/bin/env python
"""
_AddFiles_

MySQL implementation of Jobs.AddFiles
"""

import logging

from WMCore.Database.DBFormatter import DBFormatter

class AddFiles(DBFormatter):

    sql = """INSERT INTO wmbs_job_assoc (job, fileid)
               VALUES (:jobid, :fileid) """

    def getBinds(self, jobDict):
        binds = []
        for jid in jobDict:
            #For each job
            for fileID in jobDict[jid]:
                #For each file in each job
                binds.append({'jobid': jid, 'fileid': fileID})

        return binds

    def execute(self, id = None, file = None, conn = None, transaction = False, jobDict = None):

        #Adding jobDict activates bulk mode
        #Bulk mode expect jobDict of form {jid: fileid}

        if jobDict:
            binds = self.getBinds(jobDict)
        elif id and file:
            binds = DBFormatter.getBinds(self, jobid = id, fileid = file)
        else:
            logging.error('Jobs.AddFiles called with insufficient arguments')
            return

        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)
        return



#!/usr/bin/env python
# encoding: utf-8
"""
Decorators.py

Decorator Utils for GroupUser modules

Created by Dave Evans on 2010-07-20.
Copyright (c) 2010 Fermilab. All rights reserved.
"""

def requireConnection(funcRef):
    """
    _requireConnection_

    Decorator method to connect the function's class object to couch
    """
    def wrapper(self, *args, **opts):
        if not self.connected:
            self.connect()
        return funcRef(self, *args, **opts)
    return wrapper

def requireGroup(funcRef):
    """
    _requireGroup_

    Decorator method to enforce setting a group attribute

    """
    def wrapper(self, *args, **opts):
        if getattr(self, "group", None) == None:
            msg = "Group Attribute not present/set"
            raise RuntimeError(msg)
        return funcRef(self, *args, **opts)
    return wrapper


def requireUser(funcRef):
    """
    _requireUser_

    Decorator method to enforce setting a user attribute for another
    object
    """
    def wrapper(self, *args, **opts):
        if getattr(self, "owner", None) == None:
            msg = "Group Attribute not present/set"
            raise RuntimeError(msg)
        return funcRef(self, *args, **opts)
    return wrapper



#!/usr/bin/env python
"""
_DBSRESTFormatter_

A basic REST formatter. The formatter takes the data from the API call, turns it into the
appropriate format and sets the CherryPy header appropriately.

Could add YAML via http://pyyaml.org/
"""
from __future__ import (division, print_function)
from WMCore.WebTools.RESTFormatter import RESTFormatter


class DBSRESTFormatter(RESTFormatter):
    def __init__(self, config):
        RESTFormatter.__init__(self, config)
        self.supporttypes = {'application/xml': self.xml,
                             'application/atom+xml': self.atom,
                             'text/json': self.json,
                             'text/x-json': self.json,
                             'application/json': self.json,
                             'text/html': self.to_string,
                             'text/plain': self.to_string,
                             '*/*': self.json}



#! /usr/bin/env python

"""
A singleton pattern implemented in python. Adapted from ActiveState Code
Recipe 52558: The Singleton Pattern implemented with Python
http://code.activestate.com/recipes/52558/
"""




from builtins import object
class Singleton(object):
    """
    A python singleton
    """


    class SingletonImplementation(object):
        """
        Implementation of the singleton interface
        """

        def singletonId(self):
            """
            Test method, return singleton id
            """
            return id(self)


    # Storage for the instance reference
    __instance = None

    def __init__(self):
        """
        Create singleton instance
        """
        # Check whether we already have an instance
        if Singleton.__instance is None:
            # Create and remember instance
            Singleton.__instance = Singleton.SingletonImplementation()

        # Store instance reference as the only member in the handle
        self.__dict__['_Singleton__instance'] = Singleton.__instance


    def __getattr__(self, attr):
        """
        Delegate access to implementation
        """
        return getattr(self.__instance, attr)


    def __setattr__(self, attr, value):
        """
        Delegate access to implementation
        """
        return setattr(self.__instance, attr, value)



#!/usr/bin/env python
"""
_ExistsByID_

Oracle implementation of Jobs.Exists
"""

__all__ = []



from WMCore.WMBS.MySQL.Jobs.ExistsByID import ExistsByID as ExistsByIDJobMySQL

class ExistsByID(ExistsByIDJobMySQL):
    sql = ExistsByIDJobMySQL.sql



#!/usr/bin/env python
"""
_LoadFromNameAndTask_

Oracle implementation of Workflow.LoadFromNameAndTask

"""
from __future__ import division

from WMCore.WMBS.MySQL.Workflow.LoadFromNameAndTask import LoadFromNameAndTask \
    as LoadFromNameAndTaskMySQL


class LoadFromNameAndTask(LoadFromNameAndTaskMySQL):
    pass



#!/usr/bin/env python
"""
_GetTask_

MySQL implementation of Jobs.Task
"""

import logging

from WMCore.Database.DBFormatter import DBFormatter

from builtins import int

__all__ = []


class GetTask(DBFormatter):
    sql = """SELECT wmwf.task AS task, wmj.id AS jobid FROM wmbs_workflow wmwf
               INNER JOIN wmbs_subscription wms ON wmwf.id = wms.workflow
               INNER JOIN wmbs_jobgroup wmjg ON wms.id = wmjg.subscription
               INNER JOIN wmbs_job wmj ON wmjg.id = wmj.jobgroup
               WHERE wmj.id = :jobid
               """

    def format(self, results):
        """
        Should create one dictionary of the form {'jobid':'taskName'}

        """
        dictionary = self.formatDict(results)

        final = {}

        for entry in dictionary:
            final[entry['jobid']] = entry['task']

        return final

    def execute(self, jobID, conn=None, transaction=False):
        """
        Should handle bulk and regular attempts to find the task
        """

        if isinstance(jobID, list):
            if len(jobID) == 0:
                return {}
            binds = []
            for ID in jobID:
                binds.append({'jobid': int(ID)})
        elif isinstance(jobID, int):
            binds = {'jobid': int(jobID)}
        else:
            logging.error('Incompatible jobid in GetTask')
            return

        results = self.dbi.processData(self.sql, binds, conn=conn, transaction=transaction)

        return self.format(results)



"""
_CreateWMBS_

Base class for creating the WMBS database.
"""
from __future__ import print_function

import threading

from WMCore.Database.DBCreator import DBCreator
from WMCore.WMException import WMException


class CreateAgentBase(DBCreator):
    requiredTables = ["01wm_components", "02wm_workers"]

    def __init__(self, logger=None, dbi=None, params=None):
        """
        _init_

        Call the DBCreator constructor and create the list of required tables.
        """
        myThread = threading.currentThread()

        if logger is None:
            logger = myThread.logger
        if dbi is None:
            dbi = myThread.dbi

        DBCreator.__init__(self, logger, dbi)

        self.create["01wm_components"] = \
            """CREATE TABLE wm_components (
               id               INTEGER      PRIMARY KEY AUTO_INCREMENT,
               name             VARCHAR(255) NOT NULL,
               pid              INTEGER      NOT NULL,
               update_threshold INTEGER      NOT NULL,
               UNIQUE (name))"""

        self.create["02wm_workers"] = \
            """CREATE TABLE wm_workers (
               component_id  INTEGER NOT NULL,
               name          VARCHAR(255) NOT NULL,
               last_updated  INTEGER      NOT NULL,
               state         VARCHAR(255),
               pid           INTEGER,
               poll_interval INTEGER      NOT NULL,
               last_error    INTEGER,
               cycle_time    FLOAT DEFAULT 0 NOT NULL,
               outcome       VARCHAR(1000),
               error_message VARCHAR(1000),
               UNIQUE (name))"""

        self.constraints["FK_wm_component_worker"] = \
            """ALTER TABLE wm_workers ADD CONSTRAINT FK_wm_component_worker
               FOREIGN KEY(component_id) REFERENCES wm_components(id)"""

    def execute(self, conn=None, transaction=None):
        """
        _execute_

        Check to make sure that all required tables have been defined.  If
        everything is in place have the DBCreator make everything.
        """
        for requiredTable in self.requiredTables:
            if requiredTable not in self.create:
                raise WMException("The table '%s' is not defined." % \
                                  requiredTable, "WMCORE-2")

        try:
            DBCreator.execute(self, conn, transaction)
            return True
        except Exception as e:
            print("ERROR: %s" % e)
            return False



#!/usr/bin/env python
"""
_AddDupsToFileset_

Oracle implementation of Files.AddDupsToFileset
"""

from WMCore.WMBS.MySQL.Files.AddDupsToFileset import AddDupsToFileset as MySQLAddDupsToFileset

class AddDupsToFileset(MySQLAddDupsToFileset):
    sql = """INSERT INTO wmbs_fileset_files (fileid, fileset, insert_time)
               SELECT wmbs_file_details.id, :fileset, :insert_time
               FROM wmbs_file_details
               WHERE wmbs_file_details.lfn = :lfn AND NOT EXISTS
                 (SELECT lfn FROM wmbs_file_details
                    INNER JOIN wmbs_fileset_files ON
                      wmbs_file_details.id = wmbs_fileset_files.fileid
                    INNER JOIN wmbs_subscription ON
                      wmbs_fileset_files.fileset = wmbs_subscription.fileset
                    INNER JOIN wmbs_workflow ON
                      wmbs_subscription.workflow = wmbs_workflow.id
                    WHERE wmbs_file_details.lfn = :lfn AND
                          wmbs_workflow.name = :workflow)
               AND NOT EXISTS (SELECT * FROM wmbs_fileset_files
                                WHERE fileset = :fileset
                                AND insert_time = :insert_time
                                AND fileid = wmbs_file_details.id)"""

    sqlAvail = """INSERT INTO wmbs_sub_files_available (subscription, fileid)
                    SELECT wmbs_subscription.id AS subscription,
                           wmbs_file_details.id AS fileid FROM wmbs_subscription
                      INNER JOIN wmbs_file_details ON
                        wmbs_file_details.lfn = :lfn
                    WHERE wmbs_subscription.fileset = :fileset AND NOT EXISTS
                 (SELECT lfn FROM wmbs_file_details
                    INNER JOIN wmbs_fileset_files ON
                      wmbs_file_details.id = wmbs_fileset_files.fileid
                    INNER JOIN wmbs_subscription ON
                      wmbs_fileset_files.fileset = wmbs_subscription.fileset
                    INNER JOIN wmbs_workflow ON
                      wmbs_subscription.workflow = wmbs_workflow.id
                    WHERE wmbs_file_details.lfn = :lfn AND
                          wmbs_workflow.name = :workflow AND
                          wmbs_fileset_files.fileset != :fileset)
                  AND NOT EXISTS (SELECT * FROM wmbs_sub_files_available
                                    WHERE subscription = wmbs_subscription.id
                                    AND fileid = wmbs_file_details.id)"""



#!/usr/bin/env python
"""
_ScriptInvoker_

Util to invoke a Runtime Script and provide it with access to the
various bits of the job that it will need to access via the WMTaskSpace
library

This script will be invoked at runtime from the directory & subshell
environment in which the Runtime Script implementation needs to be called.


"""

from __future__ import print_function

from builtins import object
import logging
import os
import sys
import traceback

import WMCore.WMRuntime.Bootstrap as Bootstrap
from WMCore.WMRuntime.ScriptFactory import getScript


class ScriptInvoke(object):
    """
    _ScriptInvoke_

    Ctor takes two arguments:
    - module name of step module in WMTaskSpace
    - module name of the Script implementation to be invoked

    """

    def __init__(self, stepModule, scriptModule):
        self.stepModule = stepModule
        self.module = scriptModule
        self.exitCode = 0
        self.stepSpace = None
        self.script = None
        self.step = None
        self.task = None
        self.job = None
        currentDir = os.getcwd()
        Bootstrap.setupLogging(currentDir, useStdout=True)
        logging.info("Invoking scripts in current directory: %s", currentDir)

    def boot(self):
        """
        _boot_

        Import the Step Module & get the stepSpace object from it.
        Get an instance of the Script from the Script Factory

        """
        self.job = Bootstrap.loadJobDefinition()
        self.task = Bootstrap.loadTask(self.job)

        stepSpaceMod = __import__(self.stepModule,
                                  globals(), locals(), ['stepSpace'], 0)

        self.stepSpace = stepSpaceMod.stepSpace

        self.step = self.task.getStep(self.stepSpace.stepName)

        self.script = getScript(scriptModule)
        self.script.task = self.task
        self.script.step = self.step
        self.script.job = self.job
        self.script.stepSpace = self.stepSpace

    def invoke(self):
        """
        _invoke_

        call the Script implementation

        """
        self.exitCode = self.script()

    def exit(self):
        return self.exitCode


if __name__ == '__main__':

    try:
        stepModule = sys.argv[1]
        scriptModule = sys.argv[2]
    except Exception as ex:
        msg = "Usage: ScriptInvoke.py <Step Module> <Script Module>"
        raise RuntimeError(msg)

    invoker = ScriptInvoke(stepModule, scriptModule)

    try:
        invoker.boot()
    except Exception as ex:
        msg = "Error booting script invoker for step %s\n" % stepModule
        msg += "withe Script module: %s\n" % scriptModule
        msg += str(ex)
        msg += "Details:\n"
        for l in traceback.format_tb(sys.exc_info()[2]):
            msg += l
        raise RuntimeError(msg)

    try:
        invoker.invoke()
    except Exception as ex:
        msg = "Error invoking script for step %s\n" % stepModule
        msg += "withe Script module: %s\n" % scriptModule
        msg += str(ex)
        msg += "Details:\n"
        for l in traceback.format_tb(sys.exc_info()[2]):
            msg += l
        raise RuntimeError(msg)

    sys.exit(invoker.exit())



#!/usr/bin/env python

"""
_GetForJobSplittingByID_

Oracle implementation of File.GetForJobSplittingByID
"""
from WMCore.WMBS.MySQL.Files.GetForJobSplittingByID import GetForJobSplittingByID as MySQLGetByID

class GetForJobSplittingByID(MySQLGetByID):
    """
    Oracle version

    """

    sql = """SELECT id, lfn, filesize, events, first_event, merged, MIN(run) AS minrun
             FROM wmbs_file_details wfd
             LEFT OUTER JOIN wmbs_file_runlumi_map wfr ON wfr.fileid = wfd.id
             WHERE id = :fileid GROUP BY wfd.id, wfd.lfn, wfd.filesize, wfd.events, wfd.first_event,
             wfd.merged"""



#!/usr/bin/env python
"""
_New_

MySQL implementation of Jobs.New
"""

__all__ = []



import time
import logging

from WMCore.Database.DBFormatter import DBFormatter

class New(DBFormatter):
    sql = """INSERT INTO wmbs_job (jobgroup, name, state, state_time,
                                   couch_record, cache_dir, location, outcome,
                                   fwjr_path) VALUES
              (:jobgroup, :name,
               (SELECT id FROM wmbs_job_state WHERE name = 'new'),
               :state_time, :couch_record, :cache_dir,
               (SELECT id FROM wmbs_location WHERE site_name = :location),
               :outcome, :fwjr_path)"""

    getIDsql = """SELECT id as id, name as name FROM wmbs_job WHERE name= :name AND jobgroup= :jobgroup"""


    def getBinds(self, jobList):
        binds = []
        for job in jobList:
            tmpDict = {}
            tmpDict["jobgroup"]     = job.get("jobgroup")
            tmpDict["name"]         = job.get("name")
            tmpDict["couch_record"] = job.get("couch_record", None)
            tmpDict["location"]     = job.get("location", None)
            tmpDict["cache_dir"]    = job.get("cache_dir", None)
            tmpDict["state_time"]   = int(time.time())
            if job.get("outcome", 'failure') == 'success':
                tmpDict['outcome'] = 1
            else:
                tmpDict['outcome'] = 0
            tmpDict["fwjr_path"]    = job.get("fwjr", None)
            binds.append(tmpDict)

        return binds

    def format(self, input):
        result = {}
        jobList = self.formatDict(input)
        for job in jobList:
            result[job['name']] = job['id']

        return result

    def execute(self, jobgroup = None, name = None, couch_record = None, location = None, cache_dir = None,
                outcome = None, fwjr = None, conn = None, transaction = False, jobList = None):

        if outcome == None or not isinstance(outcome, str):
            outcome = 'failure'
        elif outcome.lower() == 'success':
            boolOutcome = 1
        else:
            boolOutcome = 0

        #Adding jobList enters bulk mode

        if jobList:
            binds = self.getBinds(jobList)

            self.dbi.processData(self.sql, binds, conn = conn, transaction = transaction)

            binds2 = []
            for d in binds:
                binds2.append({'name': d['name'], 'jobgroup': d['jobgroup']})

            #Now we need the IDs
            result = self.dbi.processData(self.getIDsql, binds2, conn = conn, transaction = transaction)
            return self.format(result)



        elif jobgroup and name:
            binds = {"jobgroup": jobgroup, "name": name,
                     "couch_record": couch_record, "state_time": int(time.time()),
                     "location": location, "cache_dir": cache_dir, "outcome": boolOutcome, "fwjr_path": fwjr}

            self.dbi.processData(self.sql, binds, conn = conn,
                                 transaction = transaction)
            return

        else:
            logging.error('Asked for new jobs in Jobs.New without jobgroup and name!')
            return



#!/usr/bin/env python
"""
The actual jobTracker algorithm
"""

__all__ = []

import threading
import logging
import os
import os.path

from Utils.Timers import timeFunction
from WMCore.WMExceptions import WM_JOB_ERROR_CODES
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread
from WMCore.DAOFactory import DAOFactory
from WMCore.WMException import WMException
from WMCore.FwkJobReport.Report import Report
from WMCore.JobStateMachine.ChangeState import ChangeState
from WMCore.BossAir.BossAirAPI import BossAirAPI


class JobTrackerException(WMException):
    """
    _JobTrackerException_

    A job tracker exception-handling class for the JobTracker
    """


class JobTrackerPoller(BaseWorkerThread):
    """
    _JobTrackerPoller_

    Polls the BossAir database for complete jobs
    Handles completed jobs
    """

    def __init__(self, config):
        """
        Initialise class members
        """

        BaseWorkerThread.__init__(self)
        self.config = config

        myThread = threading.currentThread()
        self.changeState = ChangeState(self.config)
        self.bossAir = BossAirAPI(config=config)
        self.daoFactory = DAOFactory(package="WMCore.WMBS",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)

        self.jobListAction = self.daoFactory(classname="Jobs.GetAllJobs")
        self.setFWJRAction = self.daoFactory(classname="Jobs.SetFWJRPath")

    def setup(self, parameters=None):
        """
        Load DB objects required for queries
        """

        return

    def terminate(self, params=None):
        """
        _terminate_

        Terminate the function after one more run.
        """
        logging.debug("terminating. doing one more pass before we die")
        self.algorithm(params)
        return

    @timeFunction
    def algorithm(self, parameters=None):
        """
        Performs the archiveJobs method, looking for each type of failure
        And deal with it as desired.
        """
        logging.info("Running Tracker algorithm")
        myThread = threading.currentThread()
        try:
            self.trackJobs()
        except WMException:
            if getattr(myThread, 'transaction', None):
                myThread.transaction.rollback()
            raise
        except Exception as ex:
            msg = "Unknown exception in JobTracker!\n"
            msg += str(ex)
            if getattr(myThread, 'transaction', None):
                myThread.transaction.rollback()
            logging.error(msg)
            raise JobTrackerException(msg)

        return

    def trackJobs(self):
        """
        _trackJobs_

        Finds a list of running jobs and the sites that they're running at,
        and passes that off to tracking.
        """
        passedJobs = []
        failedJobs = []

        jobList = self.jobListAction.execute(state="executing")
        logging.info("Have list of %i executing jobs in WMBS", len(jobList))

        if not jobList:
            return

        # retrieve completed jobs from BossAir that are 'executing' in WMBS
        completeJobs = self.bossAir.getComplete()
        logging.info("Have list of %i jobs complete in BossAir but executing in WMBS", len(completeJobs))
        logging.debug(completeJobs)

        for job in completeJobs:
            if job['id'] not in jobList:
                logging.error("Found a complete job in BossAir without a correspondent in WMBS!")
                continue
            if job['status'].lower() == 'timeout':
                failedJobs.append(job)
            else:
                passedJobs.append(job)

        # Assume all these jobs "passed" if they aren't in timeout
        self.passJobs(passedJobs)
        self.failJobs(failedJobs)

        return

    def failJobs(self, failedJobs):
        """
        _failJobs_

        Dump those jobs that have failed due to timeout
        """
        if len(failedJobs) == 0:
            return

        jrBinds = []
        for job in failedJobs:
            # Make sure the job object goes packed with fwjr_path to be persisted in couch
            jrPath = os.path.join(job.getCache(), 'Report.%i.pkl' % (job['retry_count']))
            jrBinds.append({'jobid': job['id'], 'fwjrpath': jrPath})

            fwjr = Report()
            try:
                fwjr.load(jrPath)
            except Exception:
                # Something went wrong reading the pickle
                logging.error("The pickle in %s could not be loaded, generating a new one", jrPath)
                fwjr = Report()
                fwjr.addError("NoJobReport", 99303, "NoJobReport", WM_JOB_ERROR_CODES[99303])
                fwjr.save(jrPath)
            job["fwjr"] = fwjr

        myThread = threading.currentThread()
        myThread.transaction.begin()
        self.setFWJRAction.execute(binds=jrBinds, conn=myThread.transaction.conn, transaction=True)
        self.changeState.propagate(failedJobs, 'jobfailed', 'executing')
        logging.info("Failed %i jobs", len(failedJobs))
        myThread.transaction.commit()

        return

    def passJobs(self, passedJobs):
        """
        _passJobs_

        Pass jobs and move their stuff?
        """
        if len(passedJobs) == 0:
            return

        jrBinds = []
        for job in passedJobs:
            jrPath = os.path.join(job.getCache(),
                                  'Report.%i.pkl' % (job['retry_count']))
            jrBinds.append({'jobid': job['id'], 'fwjrpath': jrPath})

        myThread = threading.currentThread()
        myThread.transaction.begin()
        self.setFWJRAction.execute(binds=jrBinds, conn=myThread.transaction.conn, transaction=True)
        self.changeState.propagate(passedJobs, 'complete', 'executing')
        myThread.transaction.commit()

        logging.info("Passed %i jobs", len(passedJobs))

        return



#!/usr/bin/env python
"""
_GetJobSlotsByCMSName_
"""

from WMCore.Database.DBFormatter import DBFormatter


class GetJobSlotsByCMSName(DBFormatter):
    """
    Get site running and pending thresholds grouped by CMS name
    """

    sql = """
        SELECT cms_name, wmbs_location_state.name as state, pending_slots, running_slots
            FROM wmbs_location
            INNER JOIN wmbs_location_state ON wmbs_location.state = wmbs_location_state.id
        """

    def formatDict(self, results):
        """
        _formatDict_

        Format the results in a dict keyed by the cms_name
        """
        formattedResults = DBFormatter.formatDict(self, results)

        dictResult = {}
        for item in formattedResults:
            site = item.pop('cms_name')
            dictResult[site] = item

        return dictResult

    def execute(self, conn=None, transaction=False, returnCursor=False):
        results = self.dbi.processData(self.sql, conn=conn, transaction=transaction)
        return self.formatDict(results)
